"""Contain the different estimators of the library."""

from abc import abstractmethod, ABC

import math
import inspect
import warnings
import platform

from sklearn.base import BaseEstimator
from sklearn.utils.validation import check_is_fitted
from sklearn.utils.extmath import safe_sparse_dot, softmax
from sklearn.exceptions import ConvergenceWarning

from collections import defaultdict

import concurrent.futures

import numpy as np
import scipy.sparse

from sklearn.utils import ClassifierTags, RegressorTags

import torch

from cyanure_gpu.data_processing import check_input_fit, check_input_inference, windows_conversion

from cyanure_gpu.logger import setup_custom_logger

from cyanure_gpu.erm.param.model_param import ModelParameters
from cyanure_gpu.erm.param.problem_param import ProblemParameters

from cyanure_gpu.erm.simple_erm import SimpleErm
from cyanure_gpu.erm.multi_erm import MultiErm

from cyanure_gpu.constants import DEVICE, TENSOR_TYPE, ARRAY_TYPE

torch.backends.cudnn.benchmark = False
np.random.seed(0)
torch.manual_seed(0)
torch.set_default_dtype(TENSOR_TYPE)
torch.set_printoptions(precision=20)

logger = setup_custom_logger("INFO")


class ERM(BaseEstimator, ABC):
    """
    The generic class for empirical risk minimization problems.

    For univariates problems, minimizes

        min_{w,b} (1/n) sum_{i=1}^n L( y_i, <w, x_i> + b)   + psi(w)

    """

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        return tags

    def _warm_start(self, X, initial_weight, nclasses):
        if self.warm_start and hasattr(self, "coef_"):
            if self.verbose:
                logger.info("Restarting with current coefficients")
            if self.fit_intercept:
                initial_weight[-1, ] = np.squeeze(self.intercept_)
                initial_weight[0:-1, ] = np.squeeze(self.coef_)
            else:
                initial_weight = np.squeeze(self.coef_)

        if self.warm_start and self.solver in ('auto', 'miso', 'catalyst-miso', 'qning-miso'):
            n = X.shape[0]
            # TODO Ecrire test pour dual surtout défensif
            reset_dual = np.any(self.dual is None)
            if not reset_dual and self._binary_problem:
                reset_dual = self.dual.shape[0] != n
            if not reset_dual and not self._binary_problem:
                reset_dual = np.any(self.dual.shape != [n, nclasses])
            if reset_dual and self.verbose:
                logger.info("Resetting dual")
            if reset_dual and self._binary_problem:
                self.dual = np.zeros(
                    n, dtype=X.dtype, order='F')
            if reset_dual and not self._binary_problem:
                self.dual = np.zeros(
                    [n, nclasses], dtype=X.dtype, order='F')

        return initial_weight

    def _initialize_weight(self, X, labels):
        nclasses = 0
        p = X.shape[1] + 1 if self.fit_intercept else X.shape[1]
        if self._binary_problem:
            initial_weight = np.zeros((p))
            yf = np.squeeze(labels)
        else:
            if labels.squeeze().ndim > 1:
                nclasses = labels.squeeze().shape[1]
                yf = np.asfortranarray(labels.T)
            else:
                nclasses = len(np.unique(labels))
                if platform.system() == "Windows":
                    yf = np.squeeze(np.intc(np.float64(labels)))
                else:
                    yf = np.squeeze(np.int32(labels))
            initial_weight = np.zeros([p, nclasses], order='F')

        initial_weight = self._warm_start(X, initial_weight, nclasses)

        initial_weight = np.asfortranarray(initial_weight, ARRAY_TYPE)

        return initial_weight, yf, nclasses

    def _multiclass_logistic(self, labels: torch.Tensor) -> str:
        if self.loss == "logistic":
            if (self.multi_class == "multinomial" or (self.multi_class == "auto" and not self._binary_problem)):
                if len(np.unique(labels)) != 2:

                    self._binary_problem = False

                    logger.info("Loss has been set to multiclass-logistic because "
                                "the multiclass parameter is set to multinomial!")

                    return "multiclass-logistic"
                else:
                    return self.loss

    def _check_max_iter(self):
        for index in range(self.n_iter_.shape[0]):
            if self.n_iter_[index] == self.max_iter:
                warnings.warn(
                    "The max_iter was reached which means the coef_ did not converge",
                    ConvergenceWarning)

    def __init__(self, loss='square', penalty='l2', fit_intercept=False, dual=None, tol=1e-3,
                 solver="auto", random_state=0, max_iter=2000, fista_restart=60,
                 verbose=True, warm_start=False, limited_memory_qning=50, multi_class="auto",
                 lambda_1=0, lambda_2=0, lambda_3=0, duality_gap_interval=5, n_threads=-1, safe=True):
        r"""
        Instantiate the ERM class.

        Parameters
        ----------
        loss: string, default='square'
            Loss function to be used. Possible choices are

                - 'square'
                    :math:`L(y,z) = \\frac{1}{2} ( y-z)^2`
                - 'logistic'
                    :math:`L(y,z) = \\log(1 + e^{-y z} )`
                - 'multiclass-logistic'
                    which is also called multinomial or softmax logistic:
                    .. math::`L(y, W^\\top x + b) = \\sum_{j=1}^k
                    \\log\\left(e^{w_j^\\top + b_j} - e^{w_y^\\top + b_y} \\right)`

        penalty (string): default='none'
            Regularization function psi. Possible choices are

            For binary_problem problems:

            - 'none'
                :math:`psi(w) = 0`
            - 'l2'
                :math:`psi(w) = \\frac{\\lambda_1}{2} ||w||_2^2`
            - 'l1'
                :math:`psi(w) = \\lambda_1 ||w||_1`

            For multivariate problems, the previous penalties operate on each
            individual (e.g., class) predictor.

            .. math::
                \\psi(W) = \\sum_{j=1}^k \\psi(w_j).

        fit_intercept (boolean): default='False'
            Learns an unregularized intercept b  (or several intercepts for
            multivariate problems)

        lambda_1 (float): default=0
            First regularization parameter

        lambda_2 (float): default=0
            Second regularization parameter, if needed

        lambda_3 (float): default=0
            Third regularization parameter, if needed

        solver (string): default='auto'
            Optimization solver. Possible choices are

            - 'ista'
            - 'ista-barzilai'
            - 'fista'
            - 'catalyst-ista'
            - 'qning-ista'  (proximal quasi-Newton method)
            - 'auto'

            see the Latex documentation for more details.
            If you are unsure, use 'auto'

        tol (float): default='1e-3'
            Tolerance parameter. For almost all combinations of loss and
            penalty functions, this parameter is based on a duality gap.
            Assuming the (non-negative) objective function is "f" and its
            optimal value is "f^*", the algorithm stops with the guarantee

            :math:`f(x_t) - f^*  <=  tol f(x_t)`

        max_iter (int): default=500
            Maximum number of iteration of the algorithm in terms of passes
            over the data

        duality_gap_interval (int): default=10
            Frequency of duality-gap computation

        verbose (boolean): default=True
            Display information or not

        n_threads (int): default=-1
            Maximum number of cores the method may use (-1 = all cores).
            Note that more cores is not always better.

        random_state (int): default=0
            Random seed

        warm_start (boolean): default=False
            Use a restart strategy

        binary_problem (boolean): default=True
            univariate or multivariate problems

        limited_memory_qning (int): default=20
            Memory parameter for the qning method

        fista_restart (int): default=50
            Restart strategy for fista (useful for computing regularization path)

        multi_class (string): default="auto"
            Determine the comportment of the instance in case of multivariate problem (for classification)

        """
        self.loss = loss
        self.penalty = penalty
        self.fit_intercept = fit_intercept
        self.dual = dual
        self.solver = solver
        self.tol = tol
        self.random_state = random_state
        self.max_iter = max_iter
        self.lambda_1 = lambda_1
        self.lambda_2 = lambda_2
        self.lambda_3 = lambda_3
        self.limited_memory_qning = limited_memory_qning
        self.fista_restart = fista_restart
        self.verbose = verbose
        self.warm_start = warm_start
        self.multi_class = multi_class
        self.duality_gap_interval = duality_gap_interval
        self.n_threads = n_threads
        self.safe = safe

    def fit(self, X, labels, le_parameter=None):
        """
        Fit the parameters.

        Parameters
        ----------
            X (numpy array):
                input n X p numpy matrix; the samples are on the rows

            y (numpy array):
                - vector of size n with real values for regression
                - vector of size n with {-1,+1} for binary classification,
                  which will be automatically converted if {0,1} are
                  provided
                - matrix of size n X k for multivariate regression
                - vector of size n with entries in {0,1,k-1} for classification
                  with k classes

        Returns
        -------
            self (ERM):
                Returns the instance
        """
        labels = np.squeeze(labels)
        initial_weight, yf, nclasses = self._initialize_weight(X, labels)

        initial_weight_torch = torch.from_numpy(initial_weight).to(DEVICE)

        if scipy.sparse.issparse(X):
            raise TypeError("Sparse data are not supported")
        else:
            training_data_fortran = np.asfortranarray(X.T, ARRAY_TYPE)

        weight_torch = torch.clone(initial_weight_torch)

        training_data_fortran, yf = windows_conversion(training_data_fortran, yf)
        labels_gpu = torch.from_numpy(yf).to(DEVICE)
        training_data_gpu = torch.from_numpy(training_data_fortran).to(DEVICE)

        loss = None
        self.le_ = le_parameter

        loss = self._multiclass_logistic(labels)

        if loss is None:
            loss = self.loss

        self.duality_gap_interval = -1 if self.duality_gap_interval <= 0 else min(self.duality_gap_interval, self.max_iter)

        model_parameter = ModelParameters(int(self.max_iter), float(self.tol), int(self.duality_gap_interval),
                                          500, 1, int(self.n_threads), int(self.limited_memory_qning),
                                          int(self.fista_restart), bool(self.verbose), False, self.solver)

        problem_parameter = ProblemParameters(float(self.lambda_1), float(self.lambda_2), float(self.lambda_3),
                                              bool(self.fit_intercept), self.penalty, loss=loss)

        optim_info = torch.empty

        if self.dual is not None:
            dual_torch = torch.from_numpy(self.dual).to(DEVICE)
        else:
            dual_torch = None

        with torch.no_grad():
            if self._binary_problem:
                erm = SimpleErm(initial_weight_torch, weight_torch, problem_parameter,
                                model_parameter, optim_info, dual_variable=dual_torch)
                self.optimization_info_, w = erm.solve_problem(training_data_gpu, labels_gpu)
            else:
                erm = MultiErm(initial_weight_torch, weight_torch, problem_parameter,
                               model_parameter, optim_info, dual_variable=dual_torch)
                if len(yf.shape) == 1:
                    self.optimization_info_, w = erm.solve_problem_vector(training_data_gpu, labels_gpu)
                else:
                    self.optimization_info_, w = erm.solve_problem_matrix(training_data_gpu, labels_gpu)

        if ("cuda" in DEVICE.type):
            w = w.cpu().numpy()
        else:
            w = w.numpy()

        self.n_iter_ = np.array([self.optimization_info_[class_index][0][-1]
                                for class_index in range(self.optimization_info_.shape[0])])

        self._check_max_iter()

        if self.fit_intercept:
            self.intercept_ = w[-1, ]
            self.coef_ = w[0:-1, ]
        else:
            self.coef_ = w

        self.n_features_in_ = self.coef_.shape[0]
        return self

    @abstractmethod
    def predict(self, X):
        """Predict the labels given an input matrix X (same format as fit)."""

    def get_weights(self):
        """
        Get the model parameters (either w or the tuple (w,b)).

        Returns
        -------
            w or (w,b) (numpy.array or tuple of numpy.array):
                Model parameters
        """
        return (self.coef_, self.intercept_) if self.fit_intercept else self.coef_

    def get_params(self, deep=True):
        """
        Get parameters for the estimator.

        Parameters
        ----------
            deep (bool, optional):
                If True returns also subobjects that are estimators. Defaults to True.

        Returns
        -------
            params (dict):
                Parameters names and values
        """
        out = {}
        for key in self._get_param_names():
            try:
                value = getattr(self, key)
            except AttributeError:
                value = None
            if deep and hasattr(value, 'get_params'):
                deep_items = value.get_params().items()
                out.update((key + '__' + k, val) for k, val in deep_items)
            out[key] = value
        return out

    @classmethod
    def _get_param_namesrestart(cls):
        init = getattr(cls.__init__, 'deprecated_original', cls.__init__)
        if init is object.__init__:
            # No explicit constructor to introspect
            return []

        # introspect the constructor arguments to find the model parameters
        # to represent
        init_signature = inspect.signature(init)
        # Consider the constructor parameters excluding 'self'
        parameters = [p for p in init_signature.parameters.values()
                      if p.name != 'self' and p.kind != p.VAR_KEYWORD]
        for p in parameters:
            if p.kind == p.VAR_POSITIONAL:
                raise RuntimeError()
        # Extract and sort argument names excluding 'self'
        return sorted([p.name for p in parameters])

    def set_params(self, **params):
        """
        Allow to change the value of parameters.

        Parameters
        ----------
            params (dict):
                Estimator parameters to set

        Raises
        ------
            ValueError:
                The parameter does not exist

        Returns
        -------
            self (ERM):
                Estimator instance
        """
        if not params:
            # Simple optimization to gain speed (inspect is slow)
            return self
        valid_params = self.get_params(deep=True)

        # Grouped by prefix
        nested_params = defaultdict(dict)
        for key, value in params.items():
            key, delim, sub_key = key.partition('__')
            if key not in valid_params:
                raise ValueError(f'Invalid parameter {key} for estimator {self}. '
                                 'Check the list of available parameters '
                                 'with `estimator.get_params().keys()`.')

            if delim:
                nested_params[key][sub_key] = value
            else:
                setattr(self, key, value)
                valid_params[key] = value

        for key, sub_params in nested_params.items():
            valid_params[key].set_params(**sub_params)
        return self

    def densify(self):
        """
        Convert coefficient matrix to dense array format.

        Converts the ``coef_`` member (back) to a numpy.ndarray. This is the
        default format of ``coef_`` and is required for fitting, so calling
        this method is only required on models that have previously been
        sparsified; otherwise, it is a no-op.

        Returns
        -------
        self (ERM):
            Fitted estimator converted to dense estimator

        """
        msg = "Estimator, %(name)s, must be fitted before densifying."
        check_is_fitted(self, msg=msg)
        if scipy.sparse.issparse(self.coef_):
            self.coef_ = self.coef_.toarray()
        return self


class ClassifierAbstraction(ERM):
    """A class to define abstract methods for classifiers."""

    @abstractmethod
    def predict_proba(self, X):
        """
        Estimate the probability for each class.

        Parameters
        ----------
            X (numpy array):
                Data matrix for which we want probabilities

        Returns
        -------
            proba (numpy.array):
                Return the probability of the samples for each class.
        """
        pass


class Regression(ERM):
    r"""
    The regression class which derives from ERM.

    The goal is to minimize the following objective:

        .. math::
            \min_{w,b} \frac{1}{n} \sum_{i=1}^n
            L\left( y_i, w^\top x_i + b\right) + \psi(w),

        where :math:`L` is a regression loss, :math:`\\psi` is a
        regularization function (or constraint), :math:`w` is a p-dimensional
        vector representing model parameters, and b is an optional
        unregularized intercept., and the targets will be real values.

    Parameters
    ----------
        loss (string): default='square'
            Loss function to be used. Possible choices are:
            Only the square loss is implemented at this point. Given two
            k-dimensional vectors y,z:

            * 'square' =>  :math:`L(y,z) = \frac{1}{2}( y-z)^2`

        penalty (string): default='none'
            Regularization function psi. Possible choices are

            For binary_problem problems:

            - 'none'
                :math:`psi(w) = 0`
            - 'l2'
                :math:`psi(w) = \frac{\lambda_1}{2} ||w||_2^2`
            - 'l1
                :math:`psi(w) = \lambda_1 ||w||_1`

            For multivariate problems, the previous penalties operate on each
            individual (e.g., class) predictor.

            .. math::
                \psi(W) = \sum_{j=1}^k \psi(w_j).

        fit_intercept (boolean): default='False'
            Learns an unregularized intercept b  (or several intercepts for
            multivariate problems)

        lambda_1 (float): default=0
            First regularization parameter

        lambda_2 (float): default=0
            Second regularization parameter, if needed

        lambda_3 (float): default=0
            Third regularization parameter, if needed

        solver (string): default='auto'
            Optimization solver. Possible choices are

            - 'ista'
            - 'ista-barzilai'
            - 'fista'
            - 'catalyst-ista'
            - 'qning-ista'  (proximal quasi-Newton method)
            - 'auto'

            see the Latex documentation for more details.
            If you are unsure, use 'auto'

        tol (float): default='1e-3'
            Tolerance parameter. For almost all combinations of loss and
            penalty functions, this parameter is based on a duality gap.
            Assuming the (non-negative) objective function is "f" and its
            optimal value is "f^*", the algorithm stops with the guarantee

            :math:`f(x_t) - f^*  <=  tol f(x_t)`

        max_iter (int): default=500
            Maximum number of iteration of the algorithm in terms of passes
            over the data

        duality_gap_interval (int): default=10
            Frequency of duality-gap computation

        verbose (boolean): default=True
            Display information or not

        n_threads (int): default=-1
            Maximum number of cores the method may use (-1 = all cores).
            Note that more cores is not always better.

        random_state (int): default=0
            Random seed

        warm_start (boolean): default=False
            Use a restart strategy

        binary_problem (boolean): default=True
            univariate or multivariate problems

        limited_memory_qning (int): default=20
            Memory parameter for the qning method

        fista_restart (int): default=50
            Restart strategy for fista (useful for computing regularization path)

    """

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.target_tags.multi_output = True
        tags.estimator_type = "regressor"
        tags.regressor_tags = RegressorTags()
        return tags

    def __init__(self, loss='square', penalty='l2', fit_intercept=True, random_state=0,
                 lambda_1=0, lambda_2=0, lambda_3=0, solver='auto', tol=1e-3,
                 duality_gap_interval=10, max_iter=500,
                 limited_memory_qning=20, fista_restart=50, verbose=True,
                 warm_start=False, n_threads=-1, dual=None, safe=True):
        super().__init__(loss=loss, penalty=penalty,
                         fit_intercept=fit_intercept, random_state=random_state, lambda_1=lambda_1,
                         lambda_2=lambda_2, lambda_3=lambda_3, solver=solver, tol=tol,
                         duality_gap_interval=duality_gap_interval, max_iter=max_iter,
                         limited_memory_qning=limited_memory_qning,
                         fista_restart=fista_restart, verbose=verbose,
                         warm_start=warm_start, n_threads=n_threads, dual=dual, safe=safe)

    def fit(self, X, y, le_parameter=None):
        """
        Fit the parameters.

        Parameters
        ----------
            X (numpy array):
                input n X p numpy matrix; the samples are on the rows

            y (numpy array):
                - vector of size n with real values for regression
                - matrix of size n X k for multivariate regression

        Returns
        -------
            self (ERM):
                Returns the instance of the class
        """
        if self.loss != 'square':
            raise ValueError("square loss should be used")

        if self.safe:
            X, labels, _ = check_input_fit(X, y, self)
        else:
            labels = y

        if labels.squeeze().ndim <= 1:
            self._binary_problem = True
        else:
            self._binary_problem = False

        return super().fit(X, labels, le_parameter)

    def predict(self, X):
        """
        Predict the labels given an input matrix X (same format as fit).

        Parameters
        ----------
            X (numpy array):
                Input matrix for the prediction

        Returns
        -------
            pred (numpy.array):
                Prediction for the X matrix
        """
        check_is_fitted(self)

        if self.safe:
            X = check_input_inference(X, self)

        X = self._validate_data(X, reset=False)

        pred = safe_sparse_dot(X, self.coef_, dense_output=False)

        if self.fit_intercept:
            pred = pred + self.intercept_

        return pred.squeeze()

    def score(self, X, y, sample_weight=None):
        r"""
        Return the coefficient of determination of the prediction.

        The coefficient of determination :math:`R^2` is defined as
        :math:`(1 - \\frac{u}{v})`, where :math:`u` is the residual
        sum of squares ``((y_true - y_pred)** 2).sum()`` and :math:`v`
        is the total sum of squares ``((y_true - y_true.mean()) ** 2).sum()``.
        The best possible score is 1.0 and it can be negative (because the
        model can be arbitrarily worse). A constant model that always predicts
        the expected value of `y`, disregarding the input features, would get
        a :math:`R^2` score of 0.0.

        Parameters
        ----------
            X (numpy array):
                Test samples.
            y (numpy.array):
                True labels for X.
            sample_weight (numpy.array, optional):
                Sample weights. Defaults to None.

        Returns
        -------
            score (float):
                :math:`R^2` of ``self.predict(X)`` wrt. `y`.
        """
        from sklearn.metrics import r2_score

        y_pred = self.predict(X)
        return r2_score(y, y_pred, sample_weight=sample_weight)


class Classifier(ClassifierAbstraction):
    r"""
    The classification class.

    The goal is to minimize the following objective:

    .. math::

        \min_{W,b} \frac{1}{n} \sum_{i=1}^n
        L\left( y_i, W^\top x_i + b\right) + \psi(W)

    where :math:`L` is a classification loss, :math:`\psi` is a regularization
    function (or constraint), :math:`W=[w_1,\ldots,w_k]` is a (p x k) matrix
    that carries the k predictors, where k is the number of classes, and
    :math:`y_i` is a label in :math:`\{1,\ldots,k\}`.
    b is a k-dimensional vector representing an unregularized intercept
    (which is optional).

    In the case of binary classification:
    :math:`w` is a p-dimensional vector representing model parameters,
    and b is an optional unregularized intercept. We expect binary labels in {-1,+1}.

    In a multivariate scenario, if you set multi_class parameter to "multinomial"
    or if multi_class="auto" and loss="logistic".
    The loss will automatically be changed to "multiclass-logistic".
    If you have a multivariate problem and are not in one of the previous configuration,
    a one-vs-all strategy will be used.


    Parameters
    ----------
    loss: string, default='square'
        Loss function to be used. Possible choices are

            - 'square'
                :math:`L(y,z) = \frac{1}{2} ( y-z)^2`
            - 'logistic'
                :math:`L(y,z) = \log(1 + e^{-y z} )`
            - 'multiclass-logistic'
                which is also called multinomial or softmax logistic:
                :math:`L(y, W^\top x + b) = \sum_{j=1}^k
                \log\left(e^{w_j^\top + b_j} - e^{w_y^\top + b_y} \right)`

    penalty (string): default='none'
        Regularization function psi. Possible choices are

        For binary_problem problems:

        - 'none'
            :math:`psi(w) = 0`
        - 'l2'
            :math:`psi(w) = \frac{\lambda_1}{2} ||w||_2^2`
        - 'l1'
            :math:`psi(w) = \lambda_1 ||w||_1`

        For multivariate problems, the previous penalties operate on each
        individual (e.g., class) predictor.

        .. math::
            \psi(W) = \sum_{j=1}^k \psi(w_j).

    fit_intercept (boolean): default='False'
        Learns an unregularized intercept b  (or several intercepts for
        multivariate problems)

    lambda_1 (float): default=0
        First regularization parameter

    lambda_2 (float): default=0
        Second regularization parameter, if needed

    lambda_3 (float): default=0
        Third regularization parameter, if needed

    solver (string): default='auto'
        Optimization solver. Possible choices are

        - 'ista'
        - 'ista-barzilai'
        - 'fista'
        - 'catalyst-ista'
        - 'qning-ista'  (proximal quasi-Newton method)
        - 'auto'

        see the Latex documentation for more details.
        If you are unsure, use 'auto'

    tol (float): default='1e-3'
        Tolerance parameter. For almost all combinations of loss and
        penalty functions, this parameter is based on a duality gap.
        Assuming the (non-negative) objective function is "f" and its
        optimal value is "f^*", the algorithm stops with the guarantee

        :math:`f(x_t) - f^*  <=  tol f(x_t)`

    max_iter (int): default=500
        Maximum number of iteration of the algorithm in terms of passes
        over the data

    duality_gap_interval (int): default=10
        Frequency of duality-gap computation

    verbose (boolean): default=True
        Display information or not

    n_threads (int): default=-1
        Maximum number of cores the method may use (-1 = all cores).
        Note that more cores is not always better.

    random_state (int): default=0
        Random seed

    warm_start (boolean): default=False
        Use a restart strategy

    binary_problem (boolean): default=True
        univariate or multivariate problems

    limited_memory_qning (int): default=20
        Memory parameter for the qning method

    fista_restart (int): default=50
        Restart strategy for fista (useful for computing regularization path)

    multi_class (string): default="auto"
        Determine the comportment of the instance in case of multivariate problem

    """

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.estimator_type = "classifier"
        tags.classifier_tags = ClassifierTags()
        return tags

    def __init__(self, loss='square', penalty='l2', fit_intercept=True, tol=1e-3, solver="auto",
                 random_state=0, max_iter=500, fista_restart=50, verbose=True,
                 warm_start=False, multi_class="auto",
                 limited_memory_qning=20, lambda_1=0, lambda_2=0, lambda_3=0,
                 duality_gap_interval=5, n_threads=-1, dual=None, safe=True):
        super().__init__(loss=loss, penalty=penalty, fit_intercept=fit_intercept,
                         tol=tol, solver=solver,
                         random_state=random_state, max_iter=max_iter, fista_restart=fista_restart,
                         verbose=verbose, warm_start=warm_start,
                         limited_memory_qning=limited_memory_qning,
                         lambda_1=lambda_1, lambda_2=lambda_2, lambda_3=lambda_3,
                         duality_gap_interval=duality_gap_interval,
                         n_threads=n_threads, multi_class=multi_class, dual=dual, safe=safe)

    def fit(self, X, y, le_parameter=None):
        """
        Fit the parameters.

        Parameters
        ----------
        X (numpy array):
            input n x p numpy matrix; the samples are on the rows

        y (numpy.array):
            Input labels.

            - vector of size n with {-1, +1} labels for binary classification,
              which will be automatically converted if labels in {0,1} are
              provided and {0,1,..., n} for multiclass classification.
        """
        if self.safe:
            X, labels, le = check_input_fit(X, y, self)
        else:
            labels = y
            le = None

        if le_parameter is not None:
            self.le_ = le_parameter
        else:
            self.le_ = le

        labels = np.squeeze(labels)
        unique = np.unique(labels)
        nb_classes = len(unique)

        if self.le_ is not None:
            self.classes_ = self.le_.classes_
        else:
            self.classes_ = unique

        if nb_classes != 2 and (nb_classes != unique.shape[0] or
                                not all(np.unique(labels) == np.arange(nb_classes))):
            logger.info("Class labels should be of the form")
            logger.info(np.arange(nb_classes))
            logger.info("but they are")
            logger.info(unique)
            logger.info(
                "The labels have been converted to respect the expected format.")

        if nb_classes == 2:
            self._binary_problem = True
            if self.le_ is not None:
                neg = labels == self.le_.transform(self.classes_)[0]
            else:
                neg = labels == self.classes_[0]
            labels = labels.astype(int)
            labels[neg] = -1
            labels[np.logical_not(neg)] = 1
        else:
            min_value = min(labels)
            if min_value != 0:
                labels = labels - min_value
            self._binary_problem = False

        super().fit(
            X, labels, le_parameter=self.le_)

        self.coef_ = self.coef_.reshape(self.coef_.shape[0], -1)
        self.coef_ = np.squeeze(self.coef_)
        if self.fit_intercept:
            self.intercept_ = self.intercept_.reshape(1, -1)

        return self

    def predict(self, X):
        """
        Predict the labels given an input matrix X (same format as fit).

        Parameters
        ----------
            X (numpy array):
                Input matrix for the prediction

        Returns
        -------
            pred (numpy.array):
                Prediction for the X matrix
        """
        check_is_fitted(self)

        pred = self.decision_function(X)

        output = None
        if len(self.classes_) == 2:
            if self.le_ is None:
                output = np.sign(pred)
                output[output == -1.0] = self.classes_[0]
                output = output.astype(np.int32)
            else:
                output = np.sign(pred)
                output[output == -1.0] = 0
                output = output.astype(np.int32)
                output = self.le_.inverse_transform(output)
        else:
            if self.le_ is None:
                output = np.argmax(pred, axis=1)
            else:
                output = self.le_.inverse_transform(np.argmax(pred, axis=1))

        return output

    def score(self, X, y):
        """
        Give an accuracy score on test data.

        Parameters
        ----------
            X (numpy array):
                Test samples.
            y (numpy.array):
                True labels for X.
            sample_weight (numpy.array, optional):
                Sample weights. Defaults to None.

        Returns
        -------
            score : float
                Mean accuracy of ``self.predict(X)`` wrt. `y`.
        """
        check_is_fitted(self)

        if self.safe:
            X = check_input_inference(X, self)

        pred = np.squeeze(self.predict(X))
        return np.sum(np.squeeze(y) == pred) / pred.shape[0]

    def decision_function(self, X):
        """
        Predict confidence scores for samples.

        Parameters
        ----------
            X (numpy array):
                The data for which we want scores

        Returns
        -------
            scores (numpy.array):
                Confidence scores per (n_samples, n_classes) combination.
                In the binary case, confidence score for self.classes_[1] where >0 means t
                his class would be predicted.
        """
        check_is_fitted(self)

        if self.safe:
            X = check_input_inference(X, self)

        if self.fit_intercept:
            scores = safe_sparse_dot(
                X, self.coef_, dense_output=False) + self.intercept_
        else:
            scores = safe_sparse_dot(X, self.coef_, dense_output=False)

        output = None
        if len(self.classes_) == 2:
            output = scores.ravel()
        else:
            output = scores.ravel() if scores.shape[1] == 1 else scores

        return output

    def predict_proba(self, X):
        """
        Estimate the probability for each class.

        Parameters
        ----------
            X (numpy array):
                Data matrix for which we want probabilities

        Returns
        -------
            proba (numpy.array):
                Return the probability of the samples for each class.
        """
        check_is_fitted(self)

        if self.safe:
            X = check_input_inference(X, self)

        decision = self.decision_function(X)
        if decision.ndim == 1:
            # Workaround for binary outcomes
            # which requires softmax prediction with only a 1D decision.
            decision = np.c_[-decision, decision]
        return softmax(decision, copy=False)


class LogisticRegression(Classifier):
    """A pre-configured class for logistic regression loss."""

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.estimator_type = "classifier"
        tags.classifier_tags = ClassifierTags()
        return tags

    def __init__(self, penalty='l2', loss='logistic', fit_intercept=True,
                 verbose=False, lambda_1=0, lambda_2=0, lambda_3=0,
                 solver='auto', tol=1e-3, duality_gap_interval=10,
                 max_iter=500, limited_memory_qning=20,
                 fista_restart=50, warm_start=False, n_threads=-1,
                 random_state=0, multi_class="auto", dual=None, safe=True):
        super().__init__(loss=loss, penalty=penalty, fit_intercept=fit_intercept,
                         solver=solver, tol=tol, random_state=random_state, verbose=verbose,
                         lambda_1=lambda_1, lambda_2=lambda_2, lambda_3=lambda_3,
                         duality_gap_interval=duality_gap_interval, max_iter=max_iter,
                         limited_memory_qning=limited_memory_qning, multi_class=multi_class,
                         fista_restart=fista_restart, warm_start=warm_start,
                         n_threads=n_threads, dual=dual, safe=safe)


def compute_r(estimator_name, aux, X, labels, active_set, fit_intercept):
    """
    Compute R coefficient corresponding to the estimator.

    Parameters
    ----------
        estimator_name (string):
            Name of the estimator class

        aux (ERM):
            Auxiliary estimator

        X (numpy array):
            Features matrix

        labels (numpy.array):
            Labels matrix

        active_set (numpy.array):
            Active set

    Returns
    -------
        R (float):
            _description_
    """
    R = None

    if len(active_set) != 0:
        pred = aux.predict(X[:, active_set])
    if estimator_name == "Lasso":
        if len(active_set) == 0:
            R = labels
        else:
            R = labels.ravel() - pred.ravel()
    elif estimator_name == "L1Logistic":
        if len(active_set) == 0:
            R = -0.5 * labels.ravel()
        else:
            R = -labels.ravel() / (1.0 + np.exp(labels.ravel() * pred.ravel()))
        if fit_intercept:
            pred += aux.intercept_

    return R


def fit_large_feature_number(estimator, aux, X, labels):
    """
    Fitting function when the number of feature is superior to 1000.

    Args
    ----
        estimator (ERM):
            Fitted estimator

        aux (ERM):
            Auxiliary estimator

        X (numpy array):
            Features matrix

        labels (numpy.array):
            Labels matrix
    """
    n, p = X.shape

    scaling = 4.0
    init = min(100, p)

    num_as = math.ceil(math.log10(p / init) / math.log10(scaling))
    active_set = []
    n_active = 0

    estimator.coef_ = np.zeros(p, dtype=X.dtype)
    estimator.restart = True

    estimator_name = type(estimator).__name__

    for ii in range(num_as):
        R = compute_r(estimator_name, aux, X, labels, active_set, estimator.fit_intercept)

        corr = np.abs(X.T @ R) / n

        if n_active > 0:
            corr[active_set] = -10e10
        n_new_as = max(
            min(init * math.ceil(scaling ** ii), p) - n_active, 0)
        new_as = corr.argsort()[-n_new_as:]

        if len(new_as) == 0 or max(corr[new_as]) <= estimator.lambda_1 * (1 + estimator.tol):
            if ii == 0:
                return estimator.fit_fallback(X, labels)
            else:
                break

        if len(active_set) > 0:
            neww = np.zeros(n_active + n_new_as,
                            dtype=X.dtype)
            neww[0:n_active] = aux.coef_
            aux.coef_ = neww
            active_set = np.concatenate((active_set, new_as))
        else:
            active_set = new_as
            aux.coef_ = np.zeros(
                len(active_set), dtype=X.dtype)

        n_active = len(active_set)
        if estimator.verbose:
            logger.info(f"Size of the active set: {n_active}")

        aux.fit(X[:, active_set], labels)

        estimator.coef_[active_set] = aux.coef_
        estimator.n_features_in_ = estimator.coef_.shape[0]
        if estimator.fit_intercept:
            estimator.intercept_ = aux.intercept_

    return estimator


def execute_fit_large_feature_number(instance, auxiliary_solver, X, labels):

    with concurrent.futures.ProcessPoolExecutor() as executor:
        f = executor.submit(fit_large_feature_number, instance, auxiliary_solver, X, labels)
        return f.result()  # will rethrow any exceptions


class Lasso(Regression):
    """
    A pre-configured class for Lasso regression.

    Using active set when the number of features is superior to 1000.
    """

    def __init__(self, lambda_1=0, solver='auto', tol=1e-3,
                 duality_gap_interval=10, max_iter=500, limited_memory_qning=20,
                 fista_restart=50, verbose=True,
                 warm_start=False, n_threads=-1, random_state=0, fit_intercept=True, dual=None, safe=True):
        super().__init__(loss='square', penalty='l1', lambda_1=lambda_1, solver=solver, tol=tol,
                         duality_gap_interval=duality_gap_interval, max_iter=max_iter,
                         limited_memory_qning=limited_memory_qning, fista_restart=fista_restart,
                         verbose=verbose, warm_start=warm_start, n_threads=n_threads,
                         random_state=random_state, fit_intercept=fit_intercept, dual=dual, safe=safe)

    def fit(self, X, y):
        """
        Fit the parameters.

        Parameters
        ----------
            X (numpy array):
                input n X p numpy matrix; the samples are on the rows

            y (numpy array):
                - vector of size n with real values for regression
                - matrix of size n X k for multivariate regression

        Returns
        -------
            self (ERM):
                Returns the instance of the class
        """
        if self.safe:
            X, labels, _ = check_input_fit(X, y, self)
        else:
            labels = y

        _, p = X.shape
        if p <= 1000:
            # no active set
            super().fit(X, labels)
        else:
            aux = Regression(loss='square', penalty='l1',
                             fit_intercept=self.fit_intercept, random_state=self.random_state,
                             lambda_1=self.lambda_1, safe=self.safe,
                             tol=self.tol, duality_gap_interval=self.duality_gap_interval,
                             max_iter=self.max_iter, solver=self.solver, verbose=self.verbose)

            estimator = execute_fit_large_feature_number(self, aux, X, labels)

            self.coef_ = estimator.coef_
            if self.fit_intercept:
                self.intercept_ = estimator.intercept_

        return self

    def fit_fallback(self, X, y):
        """
        Fit the parameters.

        Parameters
        ----------
            X (numpy array):
                input n X p numpy matrix; the samples are on the rows

            y (numpy array):
                - vector of size n with real values for regression
                - matrix of size n X k for multivariate regression

        Returns
        -------
            self (ERM):
                Returns the instance of the class
        """

        if self.safe:
            X, labels, le = check_input_fit(X, y, self)
        else:
            le = None
            labels = y
        self.le_ = le

        super().fit(X, labels, le_parameter=self.le_)

        return self


class L1Logistic(Classifier):
    """
    A pre-configured class for L1 logistic classification.

    Using active set when the number of features is superior to 1000
    """

    def __sklearn_tags__(self):
        tags = super().__sklearn_tags__()
        tags.estimator_type = "classifier"
        tags.classifier_tags = ClassifierTags()
        return tags

    def _more_tags(self):
        return {
            "_xfail_checks": {
                "check_non_transformer_estimators_n_iter": (
                    "We have a different implementation of _n_iter in the multinomial case."
                ),
            }
        }

    def __init__(self, lambda_1=0, solver='auto', tol=1e-3,
                 duality_gap_interval=10, max_iter=500, limited_memory_qning=20,
                 fista_restart=50, verbose=True, warm_start=False, n_threads=-1,
                 random_state=0, fit_intercept=True, multi_class="auto", dual=None, safe=True):
        super().__init__(loss='logistic', penalty='l1', lambda_1=lambda_1, solver=solver, tol=tol,
                         duality_gap_interval=duality_gap_interval, max_iter=max_iter,
                         limited_memory_qning=limited_memory_qning,
                         fista_restart=fista_restart, verbose=verbose,
                         warm_start=warm_start, n_threads=n_threads, random_state=random_state,
                         fit_intercept=fit_intercept, multi_class=multi_class, dual=dual, safe=safe)

    def fit(self, X, y):
        """
        Fit the parameters.

        Parameters
        ----------
        X (numpy array):
            input n x p numpy matrix; the samples are on the rows

        y (numpy.array):
            Input labels.

            - vector of size n with {-1, +1} labels for binary classification,
              which will be automatically converted if labels in {0,1} are
              provided and {0,1,..., n} for multiclass classification.
        """

        if self.multi_class == "multinomial":
            self.loss = "multiclass-logistic"
        if self.safe:
            X, labels, le = check_input_fit(X, y, self)
        else:
            le = None
            labels = y
        self.le_ = le

        _, p = X.shape
        if p <= 1000:
            # no active set
            super().fit(X, labels, le_parameter=self.le_)
        else:
            aux = Classifier(
                loss='logistic', penalty='l1', fit_intercept=self.fit_intercept,
                random_state=self.random_state,
                lambda_1=self.lambda_1, safe=self.safe, tol=self.tol,
                duality_gap_interval=self.duality_gap_interval, max_iter=self.max_iter,
                solver=self.solver, verbose=self.verbose)

            estimator = execute_fit_large_feature_number(self, aux, X, labels)

            self.coef_ = estimator.coef_
            if self.fit_intercept:
                self.intercept_ = estimator.intercept_

        return self

    def fit_fallback(self, X, y):
        """
        Fit the parameters.

        Parameters
        ----------
        X (numpy array):
            input n x p numpy matrix; the samples are on the rows

        y (numpy.array):
            Input labels.

            - vector of size n with {-1, +1} labels for binary classification,
            which will be automatically converted if labels in {0,1} are
            provided and {0,1,..., n} for multiclass classification.
        """

        if self.safe:
            X, labels, le = check_input_fit(X, y, self)
        else:
            le = None
            labels = y
        self.le_ = le

        super().fit(X, labels, le_parameter=self.le_)

        return self



import abc
import torch

from cyanure_gpu.erm.param.problem_param import ProblemParameters


class Regularizer:

    def __init__(self, model: ProblemParameters):
        self.intercept = model.intercept
        self.id = model.regul
        self.lambda_1 = model.lambda_1

    # should be able to do inplace with output=input
    @abc.abstractmethod
    def prox(self, input: torch.Tensor, eta: float) -> torch.Tensor:
        return

    @abc.abstractmethod
    def eval_tensor(self, input: torch.Tensor) -> float:
        return

    @abc.abstractmethod
    def fenchel(self, grad1: torch.Tensor, grad2: torch.Tensor) -> tuple[float, torch.Tensor, torch.Tensor]:
        return

    @abc.abstractmethod
    def print(self) -> None:
        return

    def is_lazy(self) -> bool:
        return False

    def lazy_prox(self, input: torch.Tensor, indices: torch.Tensor, eta: float) -> None:
        return None

    def provides_fenchel(self) -> bool:
        return True

    def id(self) -> str:
        return self.id

    def intercept(self) -> bool:
        return self.intercept

    def strong_convexity(self):
        return 0



import numpy as np

from cyanure_gpu.estimators import Regression, fit_large_feature_number

def test_fit_large_feature_number():
    # Test case with large number of features and dense matrix
    X = np.random.rand(100, 1001)
    labels = np.random.randint(2, size=100)
    estimator = Regression(lambda_1=0.1, fit_intercept=True, tol=1e-4, verbose=True)
    aux = Regression(lambda_1=0.1, fit_intercept=True, tol=1e-4, verbose=True)
    type(estimator).__name__ = "Lasso"
    type(aux).__name__ = "Lasso"
    result = fit_large_feature_number(estimator, aux, X, labels)
    assert isinstance(result, Regression)
    assert result.coef_.shape == (1001,)
    assert result.intercept_ is not None
    assert result.intercept_ != 0



import torch
from cyanure_gpu.erm.param.model_param import ModelParameters
from cyanure_gpu.erm.param.problem_param import ProblemParameters

from cyanure_gpu.solvers.accelerator import Catalyst, QNing
from cyanure_gpu.solvers.ista import ISTA_Solver, FISTA_Solver
from cyanure_gpu.solvers.solver import Solver

from cyanure_gpu.losses.loss import Loss
from cyanure_gpu.regularizers.regularizer import Regularizer


class Estimator:

    def __init__(self, problem_parameters: ProblemParameters, model_parameters: ModelParameters, optim_info: torch.Tensor):
        """_summary_

        Args:
            problem_parameters (ProblemParameters): _description_
            model_parameters (ModelParameters): _description_
            optim_info (torch.Tensor): _description_
        """
        self.problem_parameters = problem_parameters
        self.model_parameters = model_parameters
        self.optim_info = optim_info

    def is_loss_for_matrices(self, loss: str) -> bool:
        return loss == "SQUARE" or loss == "MULTICLASS-LOGISTIC"

    def is_regression_loss(self, loss: str) -> bool:
        return loss == "SQUARE"

    def is_regul_for_matrices(self, reg: str) -> bool:

        return reg == "L1L2" or reg == "L1LINF"

    def auto_mode(self, loss: Loss, regul: Regularizer):

        n = loss.n()
        if (n < 1000):
            solver_type = "QNING-ISTA"
        else:
            solver_type = "CATALYST-ISTA"

        return solver_type

    def get_solver(self, loss: Loss, regul: Regularizer, param: ModelParameters) -> Solver:
        solver_type = param.solver.upper()

        if "BARZILAI" in solver_type:
            linesearch = True
        else:
            linesearch = False

        solver_type = solver_type.replace('-BARZILAI', '')

        if (solver_type == "AUTO"):
            solver_type = self.auto_mode(loss, regul)
        if solver_type == "ISTA":
            solver = ISTA_Solver(loss, regul, param, linesearch)
        elif solver_type == "QNING-ISTA":
            solver = QNing(param, ISTA_Solver(loss, regul, param, linesearch))
        elif solver_type == "CATALYST-ISTA":
            solver = Catalyst(param, ISTA_Solver(loss, regul, param, linesearch))
        elif solver_type == "FISTA":
            solver = FISTA_Solver(loss, regul, param)
        else:
            solver = None
            raise NotImplementedError("This solver is not implemented !")

        return solver



import torch

from cyanure_gpu.losses.loss import Loss
from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.erm.param.model_param import ModelParameters

from cyanure_gpu.solvers.solver import Solver

from cyanure_gpu.constants import EPSILON, DEVICE

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class ISTA_Solver(Solver):

    def __init__(self, loss: Loss, regul: Regularizer, param: ModelParameters, linesearch: bool, Li: torch.Tensor = None):
        super().__init__(loss, regul, param)
        self.L = 0
        if Li is not None:
            self.Li = torch.clone(Li)
            self.L = torch.max(self.Li) / 100.0

        self.linesearch = linesearch
        if linesearch:
            self.sbb = None
            self.xbb = None

    def solver_init(self, initial_weight: torch.Tensor) -> None:
        if (self.verbose):
            self.print()
        if (self.L == 0):
            self.Li = self.loss.lipschitz_li(self.Li)
            self.L = torch.max(self.Li) / 100.0

    def solver_aux(self, weight: torch.Tensor, it: int) -> torch.Tensor:

        iter = 1
        if hasattr(self.loss, 'loss'):
            precompute = self.loss.loss.pre_compute(weight)
        else:
            precompute = self.loss.pre_compute(weight)
        fx = self.loss.eval_tensor(weight, None, precompute)
        grad = self.loss.grad(weight, None, precompute)

        alpha_max = 10e30/self.L
        alpha_min = 10e-30/self.L

        if (self.linesearch and it > 1):
            self.sbb.sub_(grad)
            self.xbb.sub_(weight)
            alpha = torch.dot(self.sbb.view(-1), self.sbb.view(-1)) / torch.norm(self.sbb)**2
            alpha = torch.min(torch.max(alpha, alpha_min), alpha_max)
            self.L = 1 / alpha

        while (iter < self.max_iter_backtracking):
            tmp2 = weight.clone()
            # Perform in-place subtraction for better performance
            tmp2 = tmp2 - grad / self.L
            tmp = self.regul.prox(tmp2, 1.0 / self.L)
            fprox = self.loss.eval_tensor(tmp)
            tmp2 = tmp.clone()
            tmp2 = tmp2 - weight
            dot_product = torch.tensordot(grad, tmp2, dims=len(grad.shape))
            norm_squared = torch.pow(torch.norm(tmp2), 2)
            if ((self.linesearch and it > 1) or fprox <= fx + dot_product + 0.5 * self.L * norm_squared + EPSILON):
                break
            self.L *= 1.5
            if (self.verbose):
                logger.info("new value for L: " + str(self.L))
            iter += 1
            if (iter == self.max_iter_backtracking):
                logger.warning("Warning: maximum number of backtracking iterations has been reached")

        if (self.linesearch):
            self.sbb = grad.clone()
            self.xbb = weight.clone()

        weight = torch.clone(tmp)

        return weight, fprox

    def print(self) -> None:
        logger.info("ISTA Solver")

    def init_kappa_acceleration(self, initial_weight: torch.Tensor) -> float:
        self.solver_init(initial_weight)
        return self.L


class FISTA_Solver(ISTA_Solver):
    def __init__(self, loss: Loss, regul: Regularizer, param: ModelParameters):
        super().__init__(loss, regul, param, linesearch=False)

    def solver_init(self, initial_weight: torch.Tensor) -> None:
        super().solver_init(initial_weight)
        self.t = torch.Tensor([1.0]).to(DEVICE)
        self.y = torch.clone(initial_weight)

    def solver_aux(self, weight: torch.Tensor, it: int = -1) -> torch.Tensor:
        # Step 1: Clone the weight tensor to avoid mutating it unintentionally.
        diff = torch.clone(weight)

        # Step 2: Call the parent class's solver_aux method and update `weight`.
        weight, _ = super().solver_aux(self.y, it)
        self.y = weight  # Update y to the new weight from super() call.

        # Step 3: Calculate the difference more robustly, adding a small epsilon.
        diff = diff - weight
        eps = torch.finfo(diff.dtype).eps  # Small epsilon to avoid very small diffs being zeroed out
        diff = torch.where(torch.abs(diff) < eps, torch.tensor(0.0, dtype=diff.dtype), diff)

        # Step 4: Update `self.t` with a stable expression.
        old_t = self.t.clone().detach()
        self.t = (1.0 + torch.sqrt(1.0 + 4.0 * torch.pow(old_t, 2))) / 2.0

        # Step 5: Perform a numerically stable update to `self.y`
        update_factor = (1.0 - old_t) / self.t
        self.y = self.y + diff * update_factor

        return weight, _

    def print(self) -> None:
        logger.info("FISTA Solver")



import torch

from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.erm.param.problem_param import ProblemParameters

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class NoRegul(Regularizer):

    def __init__(self, model: ProblemParameters):
        super().__init__(model)

        self.id = "None"

    def prox(self, input: torch.Tensor, eta: float) -> torch.Tensor:
        return input

    def eval_tensor(self, input: torch.Tensor) -> float:
        return 0

    def fenchel(self, grad1: torch.Tensor, grad2: torch.Tensor) -> float:
        return 0, grad1, grad2

    def provides_fenchel(self) -> bool:
        return False

    def print(self) -> None:
        logger.info("No regularization")



from typing import Tuple
import torch

from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.erm.param.problem_param import ProblemParameters

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class RegMat(Regularizer):

    def __init__(self, regularizer: Regularizer, model: ProblemParameters, num_cols: int, transpose: bool):
        self.transpose = transpose
        self.num_cols = num_cols.type(torch.int32)
        self.regularizer_list = list()
        for i in range(self.num_cols):
            self.regularizer_list.append(type(regularizer)(model))

    def prox(self, input: torch.Tensor, eta: float) -> torch.Tensor:
        output = torch.clone(input)
        for i in range(self.num_cols):
            if self.transpose:
                colx = input[i, :]
            else:
                colx = input[:, i]
            coly = self.regularizer_list[i].prox(colx, eta)
            if self.transpose:
                output[i, :] = coly
            else:
                output[:, i] = coly

        return output

    def eval(self, input_tensor):
        sum_value = torch.tensor(0.0)  # Initialize as a PyTorch tensor

        for i in range(self.num_cols):
            if self.transpose:
                col = input_tensor[i, :]
            else:
                col = input_tensor[:, i]
            value = self.regularizer_list[i].eval(col)
            sum_value += value

        return sum_value

    def fenchel(self, grad1: torch.Tensor, grad2: torch.Tensor) -> float:
        sum_value = torch.tensor(0.0)  # Initialize as a PyTorch tensor

        for i in range(self.num_cols):  # Assuming self.num_cols is defined somewhere
            if self.transpose:
                col1 = grad1[i, :]
                col2 = grad2[i, :]
            else:
                col1 = grad1[:, i]
                col2 = grad2[:, i]
            value = self.regularizer_list[i].fenchel(col1, col2)
            sum_value += value
            if self.transpose:
                grad1[i, :] = col1
                grad2[i, :] = col2
            else:
                grad1[:, i] = col1
                grad2[:, i] = col2

        return sum_value, grad1, grad2

    def provides_fenchel(self) -> bool:
        ok = True
        for i in range(self.num_cols):
            ok = ok and self.regularizer_list[i].provides_fenchel()

        return ok

    def print(self) -> None:
        logger.info("Regularization for matrices")
        self.regularizer_list[0].print()

    def lambda_1(self) -> float:
        return self.regularizer_list[0].lambda_1()

    def lazy_prox(self, input_tensor, indices, eta):
        output = torch.clone(input_tensor)

        for i in range(self.num_cols):
            if self.transpose:
                colx = input_tensor[i, :]
            else:
                colx = input_tensor[:, i]
            coly = self.regularizer_list[i].lazy_prox(colx, indices, eta)
            if self.transpose:
                output[i, :] = coly
            else:
                output[:, i] = coly

        return output

    def is_lazy(self) -> bool:
        return self.regularizer_list[0].is_lazy()


class RegVecToMat(Regularizer):

    def __init__(self, regularizer: Regularizer, model: ProblemParameters):
        super().__init__(model)
        parameter_tmp = model
        parameter_tmp.verbose = False
        self.regularizer = type(regularizer)(parameter_tmp)

    def prox(self, input: torch.Tensor, eta: float) -> torch.Tensor:
        weight, bias = self.get_wb(input)
        output = self.regularizer.prox(weight, eta)
        if self.intercept:
            output = torch.cat((output, torch.unsqueeze(bias, dim=1)), dim=1)
        return output

    def eval_tensor(self, input: torch.Tensor) -> float:
        weight, _ = self.get_wb(input)
        return self.regularizer.eval_tensor(torch.flatten(weight))

    def fenchel(self, grad1: torch.Tensor, grad2: torch.Tensor) -> float:
        weight, bias = self.get_wb(grad2)
        if self.intercept:
            bias_nrm_squ = torch.dot(bias, bias)
            if bias_nrm_squ > 1e-7:
                return float("inf"), grad1, grad2
        dual, grad1_flatten, weight_flatten = self.regularizer.fenchel(grad1.flatten(), weight.flatten())
        return dual, grad1_flatten.view(grad1.size()), weight_flatten.view(weight.size())

    def print(self) -> None:
        self.regularizer.print()

    def strong_convexity(self) -> float:
        return 0 if self.intercept else self.regularizer.strong_convexity()

    def lambda_1(self) -> float:
        return self.regularizer.lambda_1()

    def lazy_prox(self, input: torch.Tensor, indices: torch.Tensor, eta: float) -> None:
        weight, _ = self.get_wb(input)
        return self.regularizer.lazy_prox(weight, indices, eta)

    def is_lazy(self) -> bool:
        return self.regularizer.is_lazy()

    def get_wb(self, input: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        p = input.size(dim=1)
        if (self.intercept):
            weight = input[:, :p-1]
            bias = input[:, p-1]
        else:
            weight = input[:, :p]
            bias = None

        return weight, bias



import copy
import torch

from cyanure_gpu.erm.erm import Estimator
from cyanure_gpu.erm.simple_erm import SimpleErm
from cyanure_gpu.erm.param.problem_param import ProblemParameters
from cyanure_gpu.erm.param.model_param import ModelParameters
from cyanure_gpu.logger import setup_custom_logger
from cyanure_gpu.losses.logistic import LogisticLoss
from cyanure_gpu.losses.loss import Loss
from cyanure_gpu.losses.square_multiclass import SquareLossMat
from cyanure_gpu.regularizers.regularizer import Regularizer
from cyanure_gpu.regularizers.ridge import Ridge
from cyanure_gpu.regularizers.lasso import Lasso
from cyanure_gpu.regularizers.none import NoRegul
from cyanure_gpu.solvers.ista import ISTA_Solver
from cyanure_gpu.losses.multi_class_logistic import MultiClassLogisticLoss
from cyanure_gpu.losses.loss_matrix import LossMat
from cyanure_gpu.regularizers.regularizer_matrix import RegMat, RegVecToMat

from cyanure_gpu.constants import EPSILON, NUMBER_OPTIM_PROCESS_INFO, DEVICE

from typing import Tuple
import numpy as np

import time

logger = setup_custom_logger("INFO")


class MultiErm(Estimator):

    def __init__(self, initial_weight: torch.Tensor, weight: torch.Tensor, problem_parameters: ProblemParameters,
                 model_parameters: ModelParameters, optim_info: torch.Tensor, dual_variable: torch.Tensor):
        super().__init__(problem_parameters, model_parameters, optim_info)
        self.initial_weight = initial_weight
        self.weight = weight
        self.dual_variable = dual_variable

    # X is p x n
    # y is nclasses x n
    # W0 is p x nclasses if no intercept (or p+1 x nclasses with intercept)
    # prediction model is   W0^FeatureType X  gives  nclasses x n
    def solve_problem_vector(self, features: torch.Tensor, labels_vector: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:

        self.verify_input(features)
        nclass = torch.max(labels_vector) + 1
        loss_string = self.problem_parameters.loss.upper()

        if (super().is_regression_loss(loss_string) or not super().is_loss_for_matrices(loss_string)):
            n = labels_vector.size(dim=0)

            labels_np = np.full((int(nclass), n), fill_value=-1.0)

            # Assuming labels_vector is a PyTorch tensor
            labels_vector_np = labels_vector.cpu().numpy()

            # Set the corresponding elements to 1.0
            labels_np[labels_vector_np.astype("int32"), np.arange(n)] = 1.0

            # Convert NumPy array to PyTorch tensor
            labels = torch.tensor(labels_np.astype("float32"), device=DEVICE)

            erm_tmp = MultiErm(self.initial_weight, self.weight, self.problem_parameters,
                               self.model_parameters, self.optim_info, dual_variable=self.dual_variable)
            return erm_tmp.solve_problem_matrix(features, labels)

        if (self.model_parameters.verbose):
            logger.info("Matrix X, n=" + str(features.size(dim=1)) + ", p=" + str(features.size(dim=0)))
            pass

        loss = MultiClassLogisticLoss(features, labels_vector, self.problem_parameters.intercept)

        if (loss_string != 'MULTICLASS-LOGISTIC'):
            logger.error("Multilog loss is the only multi class implemented loss!")
            logger.info("Multilog loss is used!")

        transpose = loss.transpose()

        regul = self.get_regul_mat(nclass, transpose)

        return self.solve_mat(loss, regul)

    # X is p x n
    # y is nclasses x n
    # W0 is p x nclasses if no intercept (or p+1 x nclasses with intercept)
    # prediction model is   W0^FeatureType X  gives  nclasses x n
    def solve_problem_matrix(self, features: torch.Tensor, labels: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor]:
        self.verify_input(features)
        loss_string = self.problem_parameters.loss.upper()
        regul_string = self.problem_parameters.regul.upper()

        if (self.model_parameters.verbose):
            logger.info("Matrix X, n=" + str(features.size(dim=1)) + ", p=" + str(features.size(dim=0)))
        if (super().is_regul_for_matrices(regul_string) or super().is_loss_for_matrices(loss_string)):

            loss = self.get_loss_matrix(features, labels)

            n_class = self.initial_weight.size(dim=1)

            transpose = loss.transpose()

            regul = self.get_regul_mat(n_class, transpose)
            return self.solve_mat(loss, regul)
        else:
            self.weight = self.initial_weight
            n_class = self.initial_weight.size(dim=1)
            duality_gap_interval = max(self.model_parameters.duality_gap_interval, 1)
            self.optim_info = torch.zeros([n_class, NUMBER_OPTIM_PROCESS_INFO,
                                           int(max(self.model_parameters.max_iter / duality_gap_interval, 1))])
            parameter_tmp = self.problem_parameters
            model_parameters_tmp = copy.copy(self.model_parameters)
            model_parameters_tmp.verbose = False
            if parameter_tmp.loss == 'MULTICLASS-LOGISTIC':
                parameter_tmp.loss = 'LOGISTIC'
            initial_time = time.time()
            for ii in range(n_class):
                optim_info_col = torch.zeros([1, NUMBER_OPTIM_PROCESS_INFO,
                                              int(max(self.model_parameters.max_iter / duality_gap_interval, 1))])
                initial_weight_col = self.initial_weight[:, ii]
                weight_col = self.weight[:, ii]
                labels_col = labels[ii, :]
                dualcol = None
                if (self.dual_variable is not None and self.dual_variable.size(dim=0) == n_class):
                    dualcol = self.dual_variable[ii]
                problem_configuration = SimpleErm(initial_weight_col, weight_col, parameter_tmp,
                                                  model_parameters_tmp, optim_info_col, dualcol)
                optim_info_col, weight_col = problem_configuration.solve_problem(features, labels_col)
                if (self.dual_variable is not None and self.dual_variable.size(dim=0) == n_class):
                    self.dual_variable[ii] = problem_configuration.dual_variable

                self.weight[:, ii] = weight_col
                self.optim_info[ii, :, :optim_info_col.size(2)] = optim_info_col
                if (self.model_parameters.verbose):
                    noptim = optim_info_col.size(dim=2) - 1
                    logger.info("Solver " + str(ii) + " has terminated after " + str(optim_info_col[0, 0, noptim].cpu().numpy())
                                + " epochs in " + str(optim_info_col[0, 5, noptim].cpu().numpy()) + " seconds")
                    if (optim_info_col[0, 4, noptim] == 0):
                        logger.info("   Primal objective: " + str(optim_info_col[0, 1, noptim].cpu().numpy())
                                    + ", relative duality gap: "
                                    + str(optim_info_col[0, 3, noptim].cpu().numpy()))
                    else:
                        logger.info("   Primal objective: " + str(optim_info_col[0, 1, noptim].cpu().numpy())
                                    + ", tol: " + str(optim_info_col[0, 4, noptim].cpu().numpy()))

            final_time = time.time()
            if (self.model_parameters.verbose):
                logger.info("Time for the one-vs-all strategy")
                logger.info("Elapsed time: " + str(final_time - initial_time))

        return self.optim_info, self.weight

    def verify_input(self, X: torch.Tensor) -> None:
        if (self.problem_parameters.intercept):
            if (X.size(dim=0) + 1 != self.initial_weight.size(dim=0)):
                logger.error("Dimension of initial point is not consistent."
                             " With intercept, if X is m x n, w0 should be (n+1)-dimensional.")
                return None
        else:
            if (X.size(dim=0) != self.initial_weight.size(dim=0)):
                logger.error("Dimension of initial point is not consistent. If X is m x n, w0 should be n-dimensional.")
                return None

        if (self.model_parameters.max_iter < 0):
            raise ValueError("Maximum number of iteration must be positive")
        if (self.problem_parameters.lambda_1 < 0):
            raise ValueError("Penalty term must be positive")
        if (self.model_parameters.tol < 0):
            raise ValueError("Tolerance for stopping criteria must be positive")

    def get_loss_matrix(self, data: torch.Tensor, y: torch.Tensor) -> Loss:
        loss = None

        self.problem_parameters.loss = self.problem_parameters.loss.upper()
        if (self.problem_parameters.loss == "SQUARE"):
            loss = SquareLossMat(data, y, self.problem_parameters.intercept)
        elif (self.problem_parameters.loss == "LOGISTIC"):
            loss = LossMat(LogisticLoss(data, y, self.problem_parameters.intercept))
        else:
            logger.error("Not implemented, square loss is chosen by default")
            loss = SquareLossMat(data, y, self.problem_parameters.intercept)
        return loss

    def solve_mat(self, loss: Loss, regul: Regularizer) -> Tuple[torch.Tensor, torch.Tensor]:

        solver = None
        if (self.model_parameters.max_iter == 0):
            parameter_tmp = self.model_parameters
            parameter_tmp.verbose = False
            solver = ISTA_Solver(loss, regul, parameter_tmp, False)
            if (loss.transpose()):
                initial_weight_transposed = torch.transpose(self.initial_weight, 0, 1)
                solver.eval(initial_weight_transposed)
            else:
                solver.eval(self.initial_weight)
            self.weight = torch.clone(self.initial_weight)
        else:
            solver = self.get_solver(loss, regul, self.model_parameters)

            if (solver is None):
                self.weight = torch.clone(self.initial_weight)
                return solver.get_optim_info(), self.weight

            new_initial_weight = None
            if (self.problem_parameters.intercept):
                new_initial_weight = loss.set_intercept(self.initial_weight)
            else:
                new_initial_weight = self.initial_weight
            if (self.dual_variable is not None and self.dual_variable.size(dim=0) != 0):
                solver.set_dual_variable(self.dual_variable)

            if (loss.transpose()):
                weight_transposed = None
                initial_weight_transposed = torch.transpose(new_initial_weight, 0, 1)
                weight_transposed, fprox = solver.solve(initial_weight_transposed, weight_transposed)
                self.weight = torch.transpose(weight_transposed, 0, 1)
            else:
                self.weight, fprox = solver.solve(new_initial_weight, self.weight)

            if (self.problem_parameters.intercept):
                self.weight = loss.reverse_intercept(self.weight)

        if (self.problem_parameters.regul == "L1"):
            self.weight[abs(self.weight) < EPSILON] = 0

        return solver.get_optim_info(), self.weight

    def get_regul_mat(self, num_class: int, transpose: bool) -> Regularizer:
        regul = None

        if self.problem_parameters.regul is not None:
            self.problem_parameters.regul = self.problem_parameters.regul.upper()
            regularizer_string = self.problem_parameters.regul
            if (regularizer_string == "L2"):
                if transpose:
                    regul = RegVecToMat(Ridge(self.problem_parameters), self.problem_parameters)
                else:
                    regul = RegMat(Ridge(self.problem_parameters), self.problem_parameters, num_class, transpose)
            elif (regularizer_string == "L1"):
                if transpose:
                    regul = RegVecToMat(Lasso(self.problem_parameters), self.problem_parameters)
                else:
                    regul = RegMat(Lasso(self.problem_parameters), self.problem_parameters, num_class, transpose)
            else:
                logger.error("Not implemented, no regularization is chosen")
                regul = NoRegul(self.problem_parameters)
        else:
            regul = NoRegul(self.problem_parameters)

        return regul



import torch

from cyanure_gpu.losses.loss import ProximalPointLoss
from cyanure_gpu.erm.param.model_param import ModelParameters

from cyanure_gpu.solvers.solver import Solver

from cyanure_gpu.constants import DEVICE

from cyanure_gpu.erm.algebra import solve_binomial

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class Catalyst(Solver):
    def __init__(self, param: ModelParameters, solver: Solver):
        super().__init__(solver.loss, solver.regul, param)

        self.y = None
        self.dual_var = None
        self.count = None
        self.kappa = None
        self.alpha = None
        self.mu = None
        self.auxiliary_solver = None
        self.loss_ppa = None
        self.freq_restart = param.max_iter + 2 if solver.regul.strong_convexity() > 0 else param.freq_restart
        self.accelerated_solver = True
        self.solver = solver

    def set_dual_variable(self, initial_dual: torch.Tensor) -> None:
        self.dual_var = torch.clone(initial_dual)

    def solver_init(self, initial_weight: torch.Tensor) -> None:
        if (self.verbose):
            self.print()
        self.kappa = self.solver.init_kappa_acceleration(initial_weight)
        self.mu = self.regul.strong_convexity()
        self.count = 0
        self.accelerated_solver = self.kappa > 0  # this->_oldL/(_n) >= _mu;
        if (self.accelerated_solver):
            param2 = ModelParameters()
            param2.max_iter = 1
            param2.duality_gap_interval = 2
            param2.verbose = False
            param2.minibatch = self.minibatch

            self.solver.Li = self.solver.Li + self.kappa
            self.loss_ppa = ProximalPointLoss(self.loss, initial_weight, self.kappa)
            self.auxiliary_solver = type(self.solver)(self.loss_ppa, self.regul, param2, self.solver.linesearch, self.solver.Li)
            if (self.dual_var is not None):
                self.auxiliary_solver.set_dual_variable(self.dual_var)
            self.y = torch.clone(initial_weight)
            self.alpha = 1.0
        else:
            if (self.verbose):
                logger.info("Switching to regular solver, problem is well conditioned")
            self.solver.solver_init(initial_weight)

    def solver_aux(self, weight: torch.Tensor, it: int = -1) -> torch.Tensor:
        if (self.accelerated_solver):
            q = self.mu / (self.mu + self.kappa)
            xold = torch.clone(weight)
            weight, _ = self.auxiliary_solver.solve(self.y, weight)
            alphaold = self.alpha
            self.alpha = solve_binomial(1.0, self.alpha * self.alpha - q, -self.alpha * self.alpha)
            beta = alphaold * (1.0 - alphaold) / (alphaold * alphaold + self.alpha)
            self.count += 1
            if (self.count % self.freq_restart == 0):
                beta = 0
                self.alpha = 1.0
            self.y = torch.clone(xold)
            self.y = -beta * self.y + weight * (1.0 + beta)
            # z = anchor_point
            self.loss_ppa.anchor_point = self.y
        else:
            weight, _ = self.solver.solver_aux(weight, it)

        return weight, None

    def print(self) -> None:
        logger.info("Catalyst Accelerator")


class QNing(Catalyst):

    def __init__(self, param: ModelParameters, solver: Solver):
        super().__init__(param, solver)

        self.l_memory = param.l_memory
        self.initial_h = None
        self.ys = None
        self.ss = None
        self.rhos = None
        self.gk = None
        self.xk = None
        self.Fk = None
        self.etak = None
        self.m = None
        self.skipping_steps = 0
        self.line_search_steps = 0

    def solve(self, initial_weight: torch.Tensor, weight: torch.Tensor) -> torch.Tensor:
        weight, fprox = super().solve(initial_weight, weight)
        if (self.verbose):
            logger.info("Total additional line search steps: " + str(self.line_search_steps))
            logger.info("Total skipping l-bfgs steps: " + str(self.skipping_steps))

        return weight, fprox

    def solver_init(self, initial_weight: torch.Tensor) -> None:
        if (len(initial_weight.size()) == 2):
            dim_0_size = initial_weight.size(dim=0) * initial_weight.size(dim=1)
        else:
            dim_0_size = initial_weight.size(dim=0)
        self.ys = torch.empty(dim_0_size, self.l_memory).to(device=DEVICE, non_blocking=True)
        self.ss = torch.empty(dim_0_size, self.l_memory).to(device=DEVICE, non_blocking=True)
        self.rhos = torch.empty(self.l_memory, device=DEVICE)
        super().solver_init(initial_weight)
        if (self.accelerated_solver):
            if (self.verbose):
                logger.info("Memory parameter: " + str(self.l_memory))
            self.h0 = 1.0 / self.kappa
            self.m = 0
            self.etak = 1.0
            self.skipping_steps = 0
            self.line_search_steps = 0

    def solver_aux(self, weight: torch.Tensor, it: int = -1) -> torch.Tensor:
        if (self.accelerated_solver):
            if (self.gk is None):
                weight = self.get_gradient(weight)

            # update variable _y and test
            oldyk = torch.clone(self.y)
            oldxk = torch.clone(weight)
            oldFk = self.Fk
            oldgk = torch.clone(self.gk)
            g = self.get_lbfgs_direction()

            max_iter = 5
            self.auxiliary_solver.save_state()
            for ii in range(max_iter):
                self.y = torch.clone(oldyk)
                self.y.sub_(g, alpha=self.etak)
                self.y.add_(oldgk, alpha=((self.etak - 1.0) / self.kappa))
                weight = self.get_gradient(weight)  # _gk = kappa(x-y)
                if (self.etak == 0 or self.Fk <= (oldFk - (0.25 / self.kappa) * torch.linalg.norm(oldgk)**2)):
                    break
                if (self.Fk > 1.05 * oldFk):
                    self.auxiliary_solver.restore_state()
                    weight = torch.clone(oldxk)
                self.etak /= 2
                self.line_search_steps += 1
                if (ii == max_iter - 1 or self.etak < 0.1):
                    self.etak = 0
            if (self.Fk > 1.05 * oldFk):
                self.auxiliary_solver.restore_state()
                weight = torch.clone(oldxk)
                self.reset_lbfgs()
            else:
                oldyk = self.y - oldyk
                oldgk = self.gk - oldgk
                self.update_lbfgs_matrix(oldyk, oldgk)
            self.etak = max(min(1.0, self.etak * 1.2), 0.1)
        else:
            weight, _ = self.solver.solver_aux(weight, it)

        return weight, None

    def print(self) -> None:
        logger.info("QNing Accelerator")

    def get_lbfgs_direction(self) -> torch.Tensor:
        g = torch.clone(self.gk)
        g_vectorize = torch.flatten(g)
        g_vectorize = self.get_lbfgs_direction_aux(g_vectorize)
        g = torch.reshape(g_vectorize, self.gk.size())

        return g

    def get_lbfgs_direction_aux(self, g: torch.Tensor) -> torch.Tensor:
        # two-loop recursion algorithm
        alphas = torch.Tensor(self.l_memory).to(DEVICE)
        gamma = 1.0 / self.kappa
        for ii in range(self.m - 1, max(self.m - self.l_memory, 0) - 1, -1):
            ind = ii % self.l_memory
            cols = self.ss[:, ind]
            coly = self.ys[:, ind]
            if (ii == self.m - 1):
                gamma = torch.dot(cols, coly) / torch.linalg.norm(coly)**2
            alphas[ind] = self.rhos[ind] * torch.dot(cols, g)
            g.add_(coly, alpha=-alphas[ind])
        g = g * gamma
        for ii in range(max(self.m - self.l_memory, 0), self.m):
            ind = ii % self.l_memory
            cols = self.ss[:, ind]
            coly = self.ys[:, ind]
            beta = self.rhos[ind] * torch.dot(coly, g)
            g.add_(cols, alpha=alphas[ind] - beta)
        return g

    def update_lbfgs_matrix(self, sk: torch.Tensor, yk: torch.Tensor) -> None:

        sk_vectorize = torch.flatten(sk)
        yk_vectorize = torch.flatten(yk)

        theta = torch.dot(sk_vectorize, yk_vectorize)
        if (theta > 1e-12):
            ind = self.m % self.l_memory
            self.ys[:, ind] = yk_vectorize
            self.ss[:, ind] = sk_vectorize
            self.rhos[ind] = 1.0 / theta
            self.m += 1
        else:
            self.skipping_steps += 1

    def reset_lbfgs(self) -> None:
        self.m = 0

    def get_gradient(self, weight: torch.Tensor) -> torch.Tensor:
        self.loss_ppa.anchor_point = self.y
        weight, fprox = self.auxiliary_solver.solve(self.y, weight)
        self.gk = torch.clone(self.y)
        self.gk = torch.add(self.gk * self.kappa, weight, alpha=-self.kappa)
        self.Fk = self.loss_ppa.eval_tensor(weight, None, None, fprox) + self.regul.eval_tensor(weight)

        return weight



import abc

import torch
import time

from cyanure_gpu.erm.param.model_param import ModelParameters
from cyanure_gpu.constants import EPSILON

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class Solver:

    NUMBER_OPTIM_PROCESS_INFO = 6

    def __init__(self, loss, regul, param: ModelParameters):
        self.verbose = param.verbose
        self.duality_gap_interval = max(param.duality_gap_interval, 1)
        self.tol = param.tol
        self.max_iter = param.max_iter
        self.max_iter_backtracking = param.max_iter_backtracking
        self.best_dual = torch.tensor(-float("inf"))
        self.best_primal = torch.tensor(float("inf"))
        self.optim_info = torch.zeros((self.NUMBER_OPTIM_PROCESS_INFO, max(int(param.max_iter / self.duality_gap_interval), 1)))
        self.L = 0
        self.Li = None
        self.minibatch = param.minibatch
        self.previous_weight = None
        self.loss = loss
        self.regul = regul
        self.elapsed_time = 0
        self.initial_time = 0
        self.duality = self.loss.provides_fenchel() and self.regul.provides_fenchel()
        self.deltas = list()
        self.threshold = 1e-8
        self.old_duality_gap = 10

    def solve(self, initial_weight: torch.Tensor, weight: torch.Tensor) -> torch.Tensor:

        self.initial_time = time.time()
        weight = torch.clone(initial_weight)

        if (not self.duality and self.max_iter > 1):
            self.previous_weight = torch.clone(initial_weight)

        self.solver_init(initial_weight)
        if (self.verbose):
            logger.info("*********************************")
            self.loss.print()
            self.regul.print()
        fprox = 0
        for it in range(1, self.max_iter + 1):
            if ((it % self.duality_gap_interval) == 0):
                if (self.test_stopping_criterion(weight, it)):
                    break
            weight, fprox = self.solver_aux(weight, it)
        if (self.verbose):
            logger.info("This is the elapsed time: " + str(self.elapsed_time))
        if (self.best_primal != torch.tensor(float("inf"))):
            weight = torch.clone(self.best_weight)

        return weight, fprox

    def get_optim_info(self) -> torch.Tensor:
        count = 0
        for index in range(self.optim_info.size(dim=1)):
            if (self.optim_info[0, index] != 0):
                count += 1

        if (count > 0):
            optim = torch.zeros(1, self.NUMBER_OPTIM_PROCESS_INFO, count)
        for index in range(count):
            for inner_index in range(self.NUMBER_OPTIM_PROCESS_INFO):
                optim[0, inner_index, index] = self.optim_info[inner_index, index]

        if (count > 0):
            return optim
        else:
            return torch.unsqueeze(self.optim_info, 0)

    def eval(self, x: torch.Tensor) -> None:
        self.test_stopping_criterion(x, 1)
        self.optim_info[5, 0] = 0

    @abc.abstractmethod
    def set_dual_variable(self, initial_dual: float):
        return

    @abc.abstractmethod
    def save_state(self):
        return

    @abc.abstractmethod
    def restore_state(self):
        return

    def get_dual(self, weight: torch.Tensor) -> float:
        if (not self.regul.provides_fenchel() or not self.loss.provides_fenchel()):
            logger.error("Error: no duality gap available")
            return -float("Inf")
        grad1, grad2 = self.loss.get_dual_variable(weight)
        dual, grad1, grad2 = self.regul.fenchel(grad1, grad2)
        loss_fenchel = self.loss.fenchel(grad1)
        res = dual + loss_fenchel
        return - res

    def test_stopping_criterion(self, weight: torch.Tensor, iteration: int) -> bool:
        eval_loss = self.loss.eval_tensor(weight)
        eval_regul = self.regul.eval_tensor(weight)
        primal = eval_loss + eval_regul
        self.best_primal = torch.min(self.best_primal, primal)
        ii = max(int(iteration / self.duality_gap_interval) - 1, 0)
        optim = self.optim_info[:, ii]
        self.elapsed_time = time.time() - self.initial_time
        if (self.best_primal == primal):
            self.best_weight = torch.clone(weight)
        if (self.verbose):
            if (primal == self.best_primal):
                logger.info("Epoch: " + str(iteration) + ", primal objective: " + str(primal) +
                            ", time: " + "{:.5f}".format(self.elapsed_time))
            else:
                logger.info("Epoch: " + str(iteration) + ", primal objective: " + str(primal) +
                            ", best primal: " + str(self.best_primal) + ", time: " + "{:.5f}".format(self.elapsed_time))
        optim[0] = iteration
        optim[1] = primal
        optim[5] = self.elapsed_time
        if (self.duality):
            dual = self.get_dual(weight)
            self.best_dual = torch.max(self.best_dual, dual)
            duality_gap = (self.best_primal - self.best_dual) / torch.abs(self.best_primal)
            stop = False
            if ((iteration / self.duality_gap_interval) >= 4):
                if (all(abs(delta) < self.threshold for delta in self.deltas[-1:])):
                    stop = True
                    # TODO Add test to dtype
                    logger.warning("The solution does not improve anymore, solving has been stopped.")
            if (self.verbose):
                logger.info("Best relative duality gap: " + str(duality_gap))
            optim[2] = self.best_dual
            optim[3] = duality_gap
            if (duality_gap < self.tol):
                stop = True
            elif (duality_gap <= 0):
                logger.warning("Your problem is prone to numerical instability. It would be safer to use double.")
                stop = True
            self.optim_info[:, ii] = optim
            self.deltas.append(duality_gap - self.old_duality_gap)
            self.old_duality_gap = duality_gap
            return stop
        else:
            self.previous_weight -= weight
            diff = torch.sqrt(torch.sum(torch.pow(self.previous_weight, 2)) / max(EPSILON, torch.sum(torch.pow(weight, 2))))
            self.previous_weight = torch.clone(weight)
            optim[4] = diff
            self.optim_info[:, ii] = optim
            return diff < self.tol

    @abc.abstractmethod
    def solver_init(self, x0: torch.Tensor):
        return

    @abc.abstractmethod
    def solver_aux(self, x: torch.Tensor, it: int):
        return

    @abc.abstractmethod
    def print(self):
        return

    def minibatch(self):
        return 1



from cyanure_gpu.logger import setup_custom_logger
import torch
from cyanure_gpu.losses.loss_matrix import LinearLossMat

logger = setup_custom_logger("INFO")


class SquareLossMat(LinearLossMat):

    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool):
        super().__init__(data, y, intercept)
        self.id = "SQUARE"

    def pre_compute(self, input: torch.Tensor) -> float:

        tmp = self.pred_tensor(input, None)

        tmp = torch.sub(tmp, self.labels)

        return tmp

    def eval_tensor(self, input: torch.Tensor, matmul_result: torch.Tensor = None, precompute: torch.Tensor = None) -> float:
        if precompute is None:
            if matmul_result is not None:
                grad1 = matmul_result
            else:
                grad1 = self.pred_tensor(input, None)
            grad1 = torch.sub(grad1, self.labels)
        else:
            grad1 = precompute
        return 0.5 * torch.linalg.norm(grad1)**2 / grad1.size(dim=1)

    def eval(self, input: torch.Tensor, i: int) -> float:
        tmp = self.pred(i, input, None)
        tmp.sub_(self.labels[:, i])
        return 0.5 * torch.linalg.norm(tmp)**2

    def print(self) -> None:
        logger.info("Square Loss is used")

    def fenchel(self, input: torch.Tensor) -> float:
        num_class = input.size(dim=1)

        # Assuming 'input' and 'self.labels' are your tensors
        input_size = input.size()

        # Define the chunk size along the last dimension
        chunk_size = 1000

        # Calculate the number of chunks
        num_chunks = input_size[-1] // chunk_size

        # Initialize the result sum
        result_sum = 0.0

        # Perform batched matrix multiplication in chunks
        for i in range(num_chunks):
            start_idx = i * chunk_size
            end_idx = (i + 1) * chunk_size

            # Extract submatrices
            input_chunk = input[..., start_idx:end_idx]
            labels_chunk = self.labels[..., start_idx:end_idx]

            # Perform dot product for the chunk
            dot_product_chunk = torch.tensordot(input_chunk, labels_chunk, dims=([-1], [-1]))

            # Add the result to the sum
            result_sum += dot_product_chunk

        return 0.5 * torch.linalg.norm(input)**2 / num_class + torch.sum(torch.tensor([result_sum])) / num_class

    def get_grad_aux(self, input: torch.Tensor, matmul_result: torch.Tensor = None,
                     precompute: torch.Tensor = None) -> torch.Tensor:
        if precompute is None:
            if matmul_result is not None:
                grad1 = matmul_result
            else:
                grad1 = self.pred_tensor(input, None)
            grad1 = torch.sub(grad1, self.labels)
        else:
            grad1 = precompute
        return grad1

    def scal_grad(self, input: torch.Tensor, i: int) -> torch.Tensor:
        output = self.pred(i, input, None)
        ycol = self.labels[:, i]
        return output.sub_(ycol)

    def lipschitz_constant(self) -> float:
        return 1.0

    def get_dual_constraints(self, grad1: torch.Tensor) -> torch.Tensor:
        if (self.intercept):
            grad1 = self.center_rows(grad1)
        return grad1

    # center the matrix
    def center_rows(self, grad: torch.Tensor) -> torch.Tensor:
        mean_rows = torch.mean(grad, 0)
        grad -= mean_rows

        return grad



class ModelParameters:

    max_iter: int
    tol: float
    duality_gap_interval: int
    verbose: bool
    max_iter_backtracking: int
    minibatch: int
    threads: int
    non_uniform_sampling: bool
    l_memory: int
    freq_restart: int
    solver: str

    def __init__(self, max_iter: int = 500, tol: float = 1e-3, duality_gap_interval: int = 10, max_iter_backtracking: float = 500,
                 minibatch: int = 1, threads: int = -1, l_memory: int = 20, freq_restart: int = 50, verbose: bool = False,
                 non_uniform_sampling: bool = True, solver: str = "QNING-ISTA"):
        self.max_iter = max_iter
        self.tol = tol
        self.duality_gap_interval = duality_gap_interval
        self.max_iter_backtracking = max_iter_backtracking
        self.minibatch = minibatch
        self.threads = threads
        self.l_memory = l_memory
        self.freq_restart = freq_restart
        self.verbose = verbose
        self.non_uniform_sampling = non_uniform_sampling
        self.solver = solver



import pytest

from sklearn.utils.estimator_checks import check_estimator

from cyanure_gpu.estimators import LogisticRegression, Regression, Classifier, L1Logistic, Lasso


@pytest.mark.parametrize(
    "estimator",
    [LogisticRegression(verbose=False), Regression(verbose=False), Classifier(verbose=False), Lasso(verbose=False), L1Logistic(verbose=False)]
)
def test_all_estimators(estimator):
    return check_estimator(estimator)



import abc
import sys
import torch
import random

from typing import Tuple

from cyanure_gpu.constants import EPSILON, DEVICE, TENSOR_TYPE

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class Loss:
    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool) -> None:
        self.intercept = intercept
        self.scale_intercept = torch.tensor(1.0)
        self.input_data = data
        self.labels = y.to(TENSOR_TYPE)
        self.norms_vector = None
        self.grad_aux_compile = None

    # functions that should be implemented in derived classes (see also in protected section)
    @abc.abstractmethod
    def eval_tensor(self, input: torch.Tensor, matmul_result: torch.Tensor = None) -> float:
        return

    @abc.abstractmethod
    def eval(self, input: torch.Tensor, i: int) -> float:
        return

    # should be safe if output=input1
    @abc.abstractmethod
    def add_grad(self, input: torch.Tensor, i: int, eta: float = 1.0) -> torch.Tensor:
        return

    @abc.abstractmethod
    def scal_grad(self, input: torch.Tensor, i: int) -> float:
        return

    @abc.abstractmethod
    def add_feature_tensor(self, input: torch.Tensor, input2: torch.Tensor, s: float) -> torch.Tensor:
        return

    @abc.abstractmethod
    def add_feature(self, i: int, s: float, input2: torch.Tensor) -> torch.Tensor:
        return

    @abc.abstractmethod
    def fenchel(self, input: torch.Tensor) -> float:
        return

    @abc.abstractmethod
    def print(self) -> None:
        return

    @abc.abstractmethod
    # virtual functions that may be reimplemented, if needed
    def double_add_grad(self, input1: torch.Tensor, input2: torch.Tensor,
                        i: int, eta1: float = 1.0, eta2: float = -1.0, dummy: float = 1.0) -> torch.Tensor:
        return

    def provides_fenchel(self) -> bool:
        return True

    def norms_data(self, norms: torch.Tensor) -> torch.Tensor:
        if (self.norms_vector is None or self.norms_tensor.size(dim=0) == 0):

            self.norms_tensor = torch.pow(torch.norm(self.input_data, dim=0), 2)

            if (self.intercept):
                if norms is not None:
                    self.norms_tensor = norms + torch.pow(self.scale_intercept, 2)
                else:
                    norms = torch.pow(self.scale_intercept, 2)

        norms = self.norms_tensor

        return norms

    def norms(self, ind: int) -> float:
        return self.norms[:, ind]

    def lipschitz(self) -> float:
        norms = None
        norms = self.norms_data(norms)
        return self.lipschitz_constant()*torch.max(norms)

    def lipschitz_li(self, Li: torch.Tensor) -> float:
        Li = self.norms_data(Li)
        Li = Li * self.lipschitz_constant()
        return Li

    def transpose(self) -> bool:
        return False

    def grad(self, input: torch.Tensor, matmul_result: torch.Tensor = None, precompute: torch.Tensor = None) -> torch.Tensor:
        tmp = self.get_grad_aux(input, matmul_result, precompute)
        gradient_size = tmp.size()
        if (len(gradient_size) > 1):
            n = gradient_size[1]
        else:
            n = gradient_size[0]
        return self.add_dual_pred_tensor(tmp, input, 1.0/n, 0)

    def eval_random_minibatch(self, input: torch.Tensor, minibatch: int) -> float:
        sum = 0
        n = self.n()
        for ii in range(minibatch):
            sum += self.eval(input, random.randint(0, sys.maxsize) % n)
        return (sum/minibatch)

    def grad_random_minibatch(self, input: torch.Tensor, grad: torch.Tensor, minibatch: int) -> torch.Tensor:
        n = self.n()
        for ii in range(minibatch):
            coef = 0.0 if ii == 0 else 1.0
            grad = self.add_grad(input, random.randint(0, sys.maxsize) % n, coef)
        return grad * (1.0/minibatch)

    def kappa(self) -> float:
        return 0

    def set_anchor_point(self, z: torch.Tensor) -> None:
        pass

    def get_anchor_point(self, z: torch.Tensor) -> None:
        pass

    def get_dual_variable(self, weight: torch.Tensor, grad_aux_result: torch.Tensor = None) -> Tuple[torch.Tensor, torch.Tensor]:
        if grad_aux_result is None:
            grad1 = self.get_grad_aux(weight)
        else:
            grad1 = grad_aux_result
        grad1 = self.get_dual_constraints(grad1)
        gradient_size = grad1.size()
        if (len(gradient_size) > 1):
            n = gradient_size[1]
        else:
            n = gradient_size[0]
        return grad1, self.add_dual_pred_tensor(grad1, weight, 1.0/n, 0)

    # non-virtual function classes
    def n(self) -> int:
        return self.labels.size(0)

    def get_coordinates(self, ind: int) -> torch.Tensor:
        # TODO Wrong
        if (self.input_data.is_sparse):
            col = self.input_data[:, ind]
            indices = col

        return indices

    def set_intercept(self, initial_weight: torch.Tensor) -> torch.Tensor:
        self.scale_intercept = torch.sqrt(0.1 * torch.linalg.norm(self.input_data)**2 / self.input_data.size(dim=1))
        weight = torch.clone(initial_weight)
        if (len(weight.size()) == 1):
            weight[weight.size(0) - 1] /= self.scale_intercept
        else:
            weight[:, weight.size(1) - 1] /= self.scale_intercept

        return weight

    def reverse_intercept(self, weight: torch.Tensor) -> None:
        if (self.scale_intercept != 1.0):
            if (len(weight.size()) == 1):
                weight[weight.size(0) - 1] *= self.scale_intercept
            else:
                weight[:, weight.size(1) - 1] *= self.scale_intercept

        return weight

    def check_grad(self, input: torch.Tensor, output: torch.Tensor) -> None:
        output = torch.clone(input)
        x1 = torch.clone(input)
        x2 = torch.clone(input)
        for i in range(input.size(dim=1)):
            x1[i] -= 1e-7
            x2[i] += 1e-7
            output[i] = (self.eval_tensor(x2) - self.eval_tensor(x1)) / (2e-7)
            x1[i] += 1e-7
            x2[i] -= 1e-7

    @abc.abstractmethod
    def get_grad_aux(self, input: torch.Tensor, matmul_result: torch.Tensor = None) -> None:
        return

    @abc.abstractmethod
    def lipschitz_constant(self) -> float:
        return

    @abc.abstractmethod
    def get_dual_constraints(self, grad1: torch.Tensor) -> torch.Tensor:
        return

    @abc.abstractmethod
    def add_dual_pred_tensor(self, input: torch.Tensor, input2: torch.Tensor, a: float, b: float) -> torch.Tensor:
        return

    @abc.abstractmethod
    def add_dual_pred(self, ind: int, a: float, b: float) -> torch.Tensor:
        return

    @abc.abstractmethod
    def pred_tensor(self, input: torch.Tensor, input2: torch.Tensor) -> torch.Tensor:
        return

    @abc.abstractmethod
    def pred(self, ind: int, input: torch.Tensor, input2: torch.Tensor) -> float:
        return


class LinearLossVec(Loss):

    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool):
        super().__init__(data, y, intercept)

    def add_grad(self, input: torch.Tensor, i: int, a: float = 1.0) -> torch.Tensor:
        s = self.scal_grad(input, i, None)
        return self.add_dual_pred(i, a * s)

    def double_add_grad(self, input1: torch.Tensor, input2: torch.Tensor, i: int,
                        eta1: float = 1.0, eta2: float = -1.0, dummy: float = 1.0) -> torch.Tensor:
        res1 = self.scal_grad(input1, i, None)
        res2 = self.scal_grad(input2, i, None)
        if (res1 or res2):
            return self.add_dual_pred(i, eta1 * res1 + eta2 * res2)

    def add_feature(self, i: int, s: float, input2: torch.Tensor) -> torch.Tensor:
        return self.add_dual_pred(i, s, input2)

    def add_feature_tensor(self, input: torch.Tensor, input2: torch.Tensor, s: float) -> torch.Tensor:
        return self.add_dual_pred_tensor(input, input2, s, 1.0)

    @abc.abstractmethod
    def scal_grad(self, input: torch.Tensor, i: int) -> float:
        return

    def add_dual_pred_tensor(self, input: torch.Tensor, input2: torch.Tensor = None,
                             a: float = 1.0, b: float = 1.0) -> torch.Tensor:
        if (self.intercept):
            m = self.input_data.size(dim=0)
            output = torch.zeros((m + 1)).to(DEVICE)
            weight = self.get_w(input2)
            output[:-1] = a * torch.matmul(self.input_data, input) + weight * b
            output[-1] = self.scale_intercept * a * input.sum() + b * output[-1]
        else:
            if input2 is None:
                output = a * torch.matmul(self.input_data, input)
            else:
                output = a * torch.matmul(self.input_data, input) + b * input2
        return output

    def add_dual_pred(self, ind: int, a: float = 1.0, b: float = 1.0, input2: torch.Tensor = None) -> torch.Tensor:
        col = self.input_data[ind, :]
        m = self.input_data.size(dim=0)
        if input2 is None:
            input2 = torch.zeros((m))
        if (self.intercept):
            output = torch.zeros((m + 1))
            weight = self.get_w(input)
            col = a * weight + b * col
            output[m] = a * self.scale_intercept + b * output[m]
        else:
            output = a * col + b * input2

        return output

    def pred_tensor(self, input: torch.Tensor, input2: torch.Tensor = None) -> torch.Tensor:
        if (self.intercept):
            weight = self.get_w(input)
            output = torch.matmul(torch.transpose(self.input_data, 0, 1), weight)
            return output + (input[input.size(0) - 1] * self.scale_intercept)
        else:
            return torch.matmul(torch.transpose(self.input_data, 0, 1), input)

    def pred(self, ind: int, input: torch.Tensor, input2: torch.Tensor = None) -> float:
        col = self.input_data[ind, :]
        if (self.intercept):
            weight = self.get_w(input)
            return torch.dot(col, weight) + input[:, input.n() - 1] * self.scale_intercept
        else:
            return torch.dot(col, input)

    def get_w(self, input: torch.Tensor) -> torch.Tensor:
        n = input.size(0)
        return input[:n-1]


class ProximalPointLoss:

    def __init__(self, loss: Loss, z: torch.Tensor, kappa: float):
        self.anchor_point = z
        self.id = "PPA"
        self.kappa = kappa
        self.loss = loss

    def eval_tensor(self, input: torch.Tensor, matmul_result: torch.Tensor = None,
                    precompute: torch.Tensor = None, fprox: torch.Tensor = None) -> float:
        tmp = torch.clone(input)
        tmp = tmp - self.anchor_point
        if fprox is None:
            return self.loss.eval_tensor(input) + 0.5 * self.kappa * torch.linalg.norm(tmp)**2
        else:
            return fprox + 0.5 * self.kappa * torch.linalg.norm(tmp)**2

    def eval(self, input: torch.Tensor, i: int) -> float:
        tmp = torch.clone(input)
        tmp.sub_(self.anchor_point)
        return self.loss.eval(input, i) + 0.5 * self.kappa * torch.linalg.norm(tmp)**2

    def grad(self, input: torch.Tensor, matmul_result: torch.Tensor = None, precompute: torch.Tensor = None) -> torch.Tensor:
        grad = self.loss.grad(input, matmul_result, precompute)
        grad = grad + input * self.kappa
        grad = grad + self.anchor_point * (-self.kappa)

        return grad

    def add_grad(self, input: torch.Tensor, i: int, eta: int = 1.0) -> torch.Tensor:
        grad = self.loss.add_grad(input, i, eta)
        grad = grad + input * self.kappa * eta
        grad = grad + self.anchor_point * (-self.kappa * eta)

        return grad

    def double_add_grad(self, input1: torch.Tensor, input2: torch.Tensor,
                        i: int, eta1: int = 1.0, eta2: int = -1.0, dummy: int = 1.0) -> torch.Tensor:
        output = self.loss.double_add_grad(input1, input2, i, eta1, eta2)
        if (dummy):
            output = output + input1 * dummy * self.kappa * eta1
            output = output + input2 * dummy * self.kappa * eta2
            if (abs(eta1+eta2) > EPSILON):
                output = output + self.anchor_point * (-self.kappa * dummy * (eta1 + eta2))

        return output

    def eval_random_minibatch(self, input: torch.Tensor, minibatch: int) -> float:
        sum_value = self.loss.eval_random_minibatch(input, minibatch)
        tmp = torch.clone(input)
        tmp.sub_(self.anchor_point)
        return sum_value + 0.5 * self.kappa * torch.linalg.norm(tmp)**2

    def grad_random_minibatch(self, input: torch.Tensor, minibatch: int) -> torch.Tensor:
        grad = self.loss.grad_random_minibatch(input, minibatch)
        grad = grad + input * self.kappa
        grad = grad + self.anchor_point * (-self.kappa)

        return grad

    def print(self) -> None:
        logger.info("Proximal point loss with")
        self.loss.print()

    def provides_fenchel(self) -> bool:
        return False

    def fenchel(self, input: torch.Tensor) -> float:
        return 0

    def lipschitz(self) -> float:
        return self.loss.lipschitz() + self.kappa

    def lipschitz_li(self, Li: torch.Tensor) -> torch.Tensor:
        Li = self.loss.lipschitz_li(Li)
        return Li + self.kappa

    def scal_grad(self, input: torch.Tensor, i: int) -> float:
        return self.loss.scal_grad(input, i, None)

    def add_feature_tensor(self, input: torch.Tensor, s: float) -> torch.Tensor:
        return self.loss.add_feature_tensor(input, s)

    def add_feature(self, i: int, s: float) -> torch.Tensor:
        return self.loss.add_feature(i, s)

    def kappa(self) -> float:
        return self.kappa

    def transpose(self) -> bool:
        return self.loss.transpose()

    @abc.abstractmethod
    def get_grad_aux(self, input: torch.Tensor, matmul_result: torch.Tensor = None) -> None:
        return

    @abc.abstractmethod
    def lipschitz_constant(self) -> int:
        return

    @abc.abstractmethod
    def get_dual_constraints(self, grad1: torch.Tensor):
        return




import torch

import torch._dynamo as dynamo

import torch_tensorrt  # noqa: F401

py_loss = torch.nn.CrossEntropyLoss()

input_data = torch.randn(2048, 2448873)
loss_labels = torch.randint(0, 204, (2448873, )).to(torch.int64)
input = torch.randn(205, 2048)


def eval_tensor(input: torch.Tensor, loss_labels: torch.Tensor,
                input_data: torch.Tensor, matmul_result: torch.Tensor = None) -> float:
    if matmul_result is not None:
        tmp = matmul_result
    else:
        tmp = torch.matmul(input, input_data)

    return py_loss(tmp.T, loss_labels)


eval_tensor_compiled = torch.compile(eval_tensor, backend="dynamo", mode="max-autotune")

explanation = dynamo.explain(eval_tensor, input, loss_labels, input_data)
print(explanation)

trt_exp_program = torch.export.export(eval_tensor_compiled, args=(input, loss_labels, input_data))

torch.export.save(trt_exp_program, "trt_model.ep")



from sklearn.datasets import load_iris

import scipy.sparse as sp
from scipy import linalg, optimize, sparse

from cyanure_gpu.estimators import LogisticRegression
from cyanure_gpu.data_processing import preprocess

import numpy as np
from numpy.testing import assert_allclose, assert_almost_equal
from numpy.testing import assert_array_almost_equal, assert_array_equal, assert_array_almost_equal_nulp
import pytest

from sklearn.base import clone
from sklearn.datasets import load_iris, make_classification
from sklearn.metrics import log_loss
from sklearn.preprocessing import StandardScaler
from sklearn.utils._testing import ignore_warnings
from sklearn.utils import shuffle
from sklearn.linear_model import SGDClassifier
from sklearn.preprocessing import scale

from sklearn.exceptions import ConvergenceWarning

X = np.array([[-1, 0], [0, 1], [1, 1]]).astype("float32")
X_sp = sp.csr_matrix(X)
Y1 = np.array([0, 1, 1]).astype("float32")
Y2 = np.array([2, 1, 0]).astype("float32")
iris = load_iris()

from sklearn import preprocessing
le = preprocessing.LabelEncoder()
le.fit(iris.target_names[iris.target])

solvers = ('ista', 'fista', 'catalyst-ista', 'qning-ista', 'auto')


def check_predictions(clf, X, y):
    """Check that the model is able to fit the classification data"""
    n_samples = len(y)
    classes = np.unique(y)
    n_classes = classes.shape[0]

    predicted = clf.fit(X, y).predict(X)
    assert_array_equal(clf.classes_, classes)

    assert predicted.shape == (n_samples,)
    assert_array_equal(predicted, y)

    probabilities = clf.predict_proba(X)
    assert probabilities.shape == (n_samples, n_classes)
    assert_array_almost_equal(probabilities.sum(axis=1), np.ones(n_samples))
    assert_array_equal(probabilities.argmax(axis=1), y)


def test_predict_2_classes():
    # Simple sanity check on a 2 classes dataset
    # Make sure it predicts the correct result on simple datasets.
    check_predictions(LogisticRegression(), X, Y1)

    check_predictions(LogisticRegression(lambda_1=0.001), X, Y1)

    check_predictions(LogisticRegression(fit_intercept=False), X, Y1)


def test_error():
    # Test for appropriate exception on errors
    msg = "Penalty term must be positive"

    with pytest.raises(ValueError, match=msg):
        LogisticRegression(lambda_1=-1).fit(X, Y1)

    with pytest.raises(ValueError, match=msg):
        LogisticRegression(lambda_1="test").fit(X, Y1)

    for LR in [LogisticRegression]:
        msg = "Tolerance for stopping criteria must be positive"

        with pytest.raises(ValueError, match=msg):
            LR(tol=-1).fit(X, Y1)

        with pytest.raises(ValueError, match=msg):
            LR(tol="test").fit(X, Y1)

        msg = "Maximum number of iteration must be positive"

        with pytest.raises(ValueError, match=msg):
            LR(max_iter=-1).fit(X, Y1)

        with pytest.raises(ValueError, match=msg):
            LR(max_iter="test").fit(X, Y1)


def test_predict_3_classes():
    check_predictions(LogisticRegression(lambda_1=0.1), X, Y2)


def test_predict_iris():
    # Test logistic regression with the iris dataset
    n_samples, n_features = iris.data.shape

    target = le.transform(iris.target_names[iris.target])

    # Test that both multinomial and OvR solvers handle
    # multiclass data correctly and give good accuracy
    # score (>0.95) for the training data.
    for clf in [
        LogisticRegression(solver="ista-barzilai", lambda_1=0.1/n_samples),
        LogisticRegression(solver="qning-ista", lambda_1=0.1/n_samples, multi_class="ovr"),
        LogisticRegression(solver="qning-ista", lambda_1=0.1/n_samples),
        LogisticRegression(
            solver="ista-barzilai", tol=1e-2, random_state=42, lambda_1=0.1/n_samples
        ),
        LogisticRegression(
            solver="qning-ista",
            tol=1e-2,
            random_state=42,
            lambda_1=0.1/n_samples
        ),
    ]:
        
        clf.fit(iris.data, target)
        
        assert_array_equal(np.unique(target), clf.classes_)

        pred = clf.predict(iris.data)

        assert np.mean(pred == target) > 0.95

        probabilities = clf.predict_proba(iris.data)
        assert_array_almost_equal(probabilities.sum(axis=1), np.ones(n_samples))

        pred = probabilities.argmax(axis=1)

        assert np.mean(pred == target) > 0.95


def test_inconsistent_input():
    # Test that an exception is raised on inconsistent input
    rng = np.random.RandomState(0)
    X_ = rng.random_sample((5, 10))
    y_ = np.ones(X_.shape[0])
    y_[0] = 0

    clf = LogisticRegression(random_state=0)

    # Wrong dimensions for training data
    y_wrong = y_[:-1]

    with pytest.raises(ValueError):
        clf.fit(X, y_wrong)

    # Wrong dimensions for test data
    with pytest.raises(ValueError):
        clf.fit(X_, y_).predict(rng.random_sample((3, 12)))

def test_write_parameters():
    # Test that we can write to coef_ and intercept_
    clf = LogisticRegression(random_state=0)
    clf.fit(X, Y1)
    clf.coef_[:] = 0
    clf.intercept_[:] = 0
    assert_array_almost_equal(clf.decision_function(X), 0)


def test_nan():
    # Test proper NaN handling.
    # Regression test for Issue #252: fit used to go into an infinite loop.
    Xnan = np.array(X, dtype=np.float32)
    Xnan[0, 1] = np.nan
    logistic = LogisticRegression(random_state=0)

    with pytest.raises(ValueError):
        logistic.fit(Xnan, Y1)

def test_logistic_regression_solvers():
    X, y = make_classification(n_features=10, n_informative=5, random_state=0)

    params = dict(fit_intercept=False, random_state=42)
    ncg = LogisticRegression(solver="ista", **params)
    lib = LogisticRegression(solver="ista", **params)
    ncg.fit(X, y)
    lib.fit(X, y)
    assert_array_almost_equal(ncg.coef_, lib.coef_, decimal=2)


def test_logistic_regression_solvers_multiclass():
    X, y = make_classification(
        n_samples=200, n_features=20, n_informative=10, n_classes=3, random_state=0
    )
    params = dict(fit_intercept=False, random_state=42)
    ncg = LogisticRegression(solver="ista", **params)
    lib = LogisticRegression(solver="ista", **params)
    ncg.fit(X, y)
    lib.fit(X, y)
    assert_array_almost_equal(ncg.coef_, lib.coef_, decimal=2)


@pytest.mark.parametrize("solver", solvers)
@pytest.mark.xfail(reason="Ill-conditionned problem.")
def test_logistic_regression_multinomial(solver):
    # Tests for the multinomial option in logistic regression

    # Some basic attributes of Logistic Regression
    n_samples, n_features, n_classes = 50, 20, 3
    X, y = make_classification(
        n_samples=n_samples,
        n_features=n_features,
        n_informative=10,
        n_classes=n_classes,
        random_state=0,
    )

    X = StandardScaler(with_mean=False).fit_transform(X)    
    # 'qning-svrg' is used as a referenced
    solver_ref = "qning-svrg"
    ref_i = LogisticRegression(solver=solver_ref, multi_class="multinomial")
    ref_w = LogisticRegression(
        solver=solver_ref, multi_class="multinomial", fit_intercept=False)
    ref_i.fit(X, y)
    ref_w.fit(X, y)
    assert ref_i.coef_.shape == (n_features, n_classes)
    assert ref_w.coef_.shape == (n_features, n_classes)
    clf_i = LogisticRegression(
        solver=solver,
        multi_class="multinomial",
        random_state=42,
        max_iter=2000,
        tol=1e-7
    )
    clf_w = LogisticRegression(
        solver=solver,
        multi_class="multinomial",
        random_state=42,
        max_iter=2000,
        tol=1e-7,
        fit_intercept=False
    )
    clf_i.fit(X, y)
    clf_w.fit(X, y)
    assert clf_i.coef_.shape == (n_features, n_classes)
    assert clf_w.coef_.shape == (n_features, n_classes)

    # Compare solutions between lbfgs and the other solvers
    assert_allclose(ref_i.coef_, clf_i.coef_, rtol=1e-2)
    assert_allclose(ref_w.coef_, clf_w.coef_, rtol=1e-2)
    assert_allclose(ref_i.intercept_, clf_i.intercept_, rtol=1e-2)


def test_liblinear_decision_function_zero():
    # Test negative prediction when decision_function values are zero.
    # Liblinear predicts the positive class when decision_function values
    # are zero. This is a test to verify that we do not do the same.
    # See Issue: https://github.com/scikit-learn/scikit-learn/issues/3600
    # and the PR https://github.com/scikit-learn/scikit-learn/pull/3623
    X, y = make_classification(n_samples=5, n_features=5, random_state=0)
    clf = LogisticRegression(fit_intercept=False, solver="ista")
    clf.fit(X, y)

    # Dummy data such that the decision function becomes zero.
    X = np.zeros((5, 5))
    assert_array_equal(clf.predict(X), np.zeros(5))

def test_logreg_l1():
    rng = np.random.RandomState(42)
    n_samples = 50
    X, y = make_classification(n_samples=n_samples, n_features=25, random_state=0)
    X_noise = rng.normal(size=(n_samples, 3))
    X_constant = np.ones(shape=(n_samples, 2))
    X = np.concatenate((X, X_noise, X_constant), axis=1)

    lr_liblinear = LogisticRegression(
        penalty="l1",
        solver="qning-ista",
        fit_intercept=True,
        max_iter=500,
        multi_class="ovr",
        lambda_1=1/(2*n_samples)
    )
    lr_liblinear.fit(X, y)

    lr_saga = LogisticRegression(
        penalty="l1",
        solver="catalyst-ista",
        fit_intercept=True,
        max_iter=500,
        multi_class="ovr",
        lambda_1=1/(2*n_samples)
    )
    lr_saga.fit(X, y)
    
    assert_allclose(lr_liblinear.coef_, lr_saga.coef_, rtol=0.02)

    # Noise and constant features should be regularized to zero by the l1
    # penalty
    assert_array_almost_equal(lr_liblinear.coef_[-5:], np.zeros(5))
    assert_array_almost_equal(lr_saga.coef_[-5:], np.zeros(5))


def test_logreg_predict_proba_multinomial():
    X, y = make_classification(
        n_samples=10, n_features=50, random_state=0, n_classes=3, n_informative=30
    )

    # Predicted probabilities using the true-entropy loss should give a
    # smaller loss than those using the ovr method.
    clf_multi = LogisticRegression(multi_class="multinomial", solver="qning-ista", lambda_1=0)
    clf_multi.fit(X, y)
    clf_multi_loss = log_loss(y, clf_multi.predict_proba(X))
    clf_ovr = LogisticRegression(multi_class="ovr", solver="qning-ista", lambda_1=0)
    clf_ovr.fit(X, y)
    clf_ovr_loss = log_loss(y, clf_ovr.predict_proba(X))
    np.testing.assert_almost_equal(clf_ovr_loss, clf_multi_loss, decimal=6)


@pytest.mark.parametrize("max_iter", np.arange(1, 5))
@pytest.mark.parametrize(
    "solver, message",
    [
        (
            "ista",
            "The max_iter was reached which means the coef_ did not converge",
        ),
        (
            "fista",
            "The max_iter was reached which means the coef_ did not converge",
        ),
        ("catalyst-ista", "The max_iter was reached which means the coef_ did not converge"),
        ("qning-ista", "The max_iter was reached which means the coef_ did not converge"),
    ],
)
def test_max_iter(max_iter, solver, message):
    # Test that the maximum number of iteration is reached
    X, y_bin = iris.data, iris.target.copy()
    y_bin[y_bin == 2] = 0

    lr = LogisticRegression(
        max_iter=max_iter,
        tol=1e-15,
        random_state=0,
        solver=solver,
    )
    with pytest.warns(ConvergenceWarning, match=message):
        lr.fit(X, y_bin)

    assert lr.n_iter_[0] == max_iter


@pytest.mark.parametrize("solver", solvers)
def test_n_iter(solver):
    # Test that self.n_iter_ has the correct format.
    X, y = iris.data, iris.target

    y_bin = y.copy()
    y_bin[y_bin == 2] = 0

    # OvR case
    n_classes = np.unique(y).shape[0]
    clf = LogisticRegression(
        tol=1e-2, solver=solver, lambda_1=1.0, random_state=42, multi_class="ovr",
    )
    clf.fit(X, y)
    assert clf.n_iter_.shape == (n_classes,)

    # multinomial case

@pytest.mark.parametrize("solver", solvers)
@pytest.mark.parametrize("warm_start", (True, False))
@pytest.mark.parametrize("fit_intercept", (True, False))
@pytest.mark.parametrize("multi_class", ["ovr", "multinomial"])
def test_warm_start(solver, warm_start, fit_intercept, multi_class):
    # A 1-iteration second fit on same data should give almost same result
    # with warm starting, and quite different result without warm starting.
    # Warm starting does not work with liblinear solver.
    X, y = iris.data, iris.target

    clf = LogisticRegression(
        tol=1e-4,
        multi_class=multi_class,
        warm_start=warm_start,
        solver=solver,
        random_state=42,
        fit_intercept=fit_intercept,
        penalty="l2",
        lambda_1=1/X.shape[0],
        max_iter=500
    )
    with ignore_warnings(category=ConvergenceWarning):
        clf.fit(X, y)
        coef_1 = clf.coef_
        import sys

        clf.max_iter = 1
        clf.fit(X, y)
    cum_diff = np.sum(np.abs(coef_1 - clf.coef_))
    msg = (
        "Warm starting issue with %s solver in %s mode "
        "with fit_intercept=%s and warm_start=%s"
        % (solver, multi_class, str(fit_intercept), str(warm_start))
    )
    if warm_start:
        assert 2.0 > cum_diff, msg
    else:
        assert cum_diff > 2.0, msg

# alpha=1e-3 is time consumingTest 
@pytest.mark.parametrize("penalty", ["l1", "l2"])
@pytest.mark.parametrize("alpha", np.logspace(-1, 1, 3))
@pytest.mark.xfail(reason="Solvers without acceleration fail without regularization.")
def test_ista_vs_svrg_fail(penalty, alpha):
    iris = load_iris()
    X, y = iris.data, iris.target
    X = np.concatenate([X] * 3)
    y = np.concatenate([y] * 3)

    X_bin = X[y <= 1]
    y_bin = y[y <= 1] * 2 - 1

    X_sparse, y_sparse = make_classification(
        n_samples=50, n_features=20, random_state=0
    )
    X_sparse = sparse.csr_matrix(X_sparse)

    for (X, y) in ((X_bin, y_bin), (X_sparse, y_sparse)):
        n_samples = X.shape[0]
        saga = LogisticRegression(
            lambda_1=1 / (n_samples * alpha),
            solver="ista",
            max_iter=1000,
            fit_intercept=True,
            penalty=penalty,
            random_state=0,
            tol=1e-12,
            n_threads=-1,
        )

        liblinear = LogisticRegression(
            lambda_1=1 / (n_samples * alpha),
            solver="svrg",
            max_iter=1000,
            fit_intercept=True,
            penalty=penalty,
            random_state=0,
            tol=1e-12,
            n_threads=-1,
        )

        saga.fit(X, y)
        liblinear.fit(X, y)
        # Convergence for alpha=1e-3 is very slow
        assert_array_almost_equal(saga.coef_, liblinear.coef_, 2)

@pytest.mark.parametrize("solver", ['qning-ista', 'qning-svrg', 'qning-ista', 'auto'])
@pytest.mark.parametrize("fit_intercept", [False, True])
@pytest.mark.xfail(reason="There is an infinity of solution.")
def test_dtype_match_regul_fail(solver, fit_intercept):
    # Test that np.float32 input data is not cast to np.float32 when possible
    # and that the output is approximately the same no matter the input format.

    out32_type = np.float32

    X_32 = np.array(X).astype(np.float32)
    y_32 = np.array(Y1).astype(np.float32)
    X_32 = np.array(X).astype(np.float32)
    y_32 = np.array(Y1).astype(np.float32)
    solver_tol = 5e-4

    lr_templ = LogisticRegression(
        solver=solver,
        random_state=42,
        tol=solver_tol,
        fit_intercept=fit_intercept,
    )

    # Check 32-bit type consistency
    lr_32 = clone(lr_templ)
    lr_32.fit(X_32, y_32)
    assert lr_32.coef_.dtype == out32_type

    # Check 32-bit type consistency
    lr_32 = clone(lr_templ)
    lr_32.fit(X_32, y_32)
    assert lr_32.coef_.dtype == np.float32

    # solver_tol bounds the norm of the loss gradient
    # dw ~= inv(H)*grad ==> |dw| ~= |inv(H)| * solver_tol, where H - hessian
    #
    # See https://github.com/scikit-learn/scikit-learn/pull/13325
    #
    # with  Z = np.hstack((np.ones((3,1)), np.array(X)))
    # In [8]: np.linalg.norm(np.diag([0,2,2]) + np.linalg.inv((Z.T @ Z)/4))
    # Out[8]: 1.7193336918135917

    # factor of 2 to get the ball diameter
    atol = 2 * 1.72 * solver_tol
    if True: # os.name == "nt" and _IS_32BIT:
        # FIXME from scikit-learn test
        atol = 1e-1

    # Check accuracy consistency
    assert_allclose(lr_32.coef_, lr_32.coef_.astype(np.float32), atol=atol)

@pytest.mark.parametrize("solver", ['catalyst-ista', 'fista'])
@pytest.mark.parametrize("fit_intercept", [False, True])
def test_dtype_match_regul(solver, fit_intercept):
    # Test that np.float32 input data is not cast to np.float32 when possible
    # and that the output is approximately the same no matter the input format.

    out32_type = np.float32

    X_32 = np.array(X).astype(np.float32)
    y_32 = np.array(Y1).astype(np.float32)
    X_32 = np.array(X).astype(np.float32)
    y_32 = np.array(Y1).astype(np.float32)
    solver_tol = 5e-4

    lr_templ = LogisticRegression(
        solver=solver,
        random_state=42,
        tol=solver_tol,
        fit_intercept=fit_intercept,
        penalty='l2',
        lambda_1=0.1
    )

    # Check 32-bit type consistency
    lr_32 = clone(lr_templ)
    lr_32.fit(X_32, y_32)
    assert lr_32.coef_.dtype == out32_type

    # Check 32-bit type consistency
    lr_32 = clone(lr_templ)
    lr_32.fit(X_32, y_32)
    assert lr_32.coef_.dtype == np.float32

    # solver_tol bounds the norm of the loss gradient
    # dw ~= inv(H)*grad ==> |dw| ~= |inv(H)| * solver_tol, where H - hessian
    #
    # See https://github.com/scikit-learn/scikit-learn/pull/13325
    #
    # with  Z = np.hstack((np.ones((3,1)), np.array(X)))
    # In [8]: np.linalg.norm(np.diag([0,2,2]) + np.linalg.inv((Z.T @ Z)/4))
    # Out[8]: 1.7193336918135917

    # factor of 2 to get the ball diameter
    atol = 2 * 1.72 * solver_tol
    if True: # os.name == "nt" and _IS_32BIT:
        # FIXME from scikit-learn test
        atol = 1e-1

    # Check accuracy consistency
    assert_allclose(lr_32.coef_, lr_32.coef_.astype(np.float32), atol=atol)

@pytest.mark.parametrize("solver", solvers)
@pytest.mark.parametrize("fit_intercept", [False, True])
def test_dtype_match(solver, fit_intercept):
    # Test that np.float32 input data is not cast to np.float32 when possible
    # and that the output is approximately the same no matter the input format.

    if solver != "catalyst-ista" and solver != 'fista' and solver != 'auto' and 'qning' not in solver:
        out32_type = np.float32

        X_32 = np.array(X).astype(np.float32)
        y_32 = np.array(Y1).astype(np.float32)
        X_32 = np.array(X).astype(np.float32)
        y_32 = np.array(Y1).astype(np.float32)
        solver_tol = 5e-4

        lr_templ = LogisticRegression(
            solver=solver,
            random_state=42,
            tol=solver_tol,
            fit_intercept=fit_intercept,
        )

        # Check 32-bit type consistency
        lr_32 = clone(lr_templ)
        lr_32.fit(X_32, y_32)
        assert lr_32.coef_.dtype == out32_type

        # Check 32-bit type consistency
        lr_32 = clone(lr_templ)
        lr_32.fit(X_32, y_32)
        assert lr_32.coef_.dtype == np.float32

        # solver_tol bounds the norm of the loss gradient
        # dw ~= inv(H)*grad ==> |dw| ~= |inv(H)| * solver_tol, where H - hessian
        #
        # See https://github.com/scikit-learn/scikit-learn/pull/13325
        #
        # with  Z = np.hstack((np.ones((3,1)), np.array(X)))
        # In [8]: np.linalg.norm(np.diag([0,2,2]) + np.linalg.inv((Z.T @ Z)/4))
        # Out[8]: 1.7193336918135917

        # factor of 2 to get the ball diameter
        atol = 2 * 1.72 * solver_tol
        if True: # os.name == "nt" and _IS_32BIT:
            # FIXME from scikit-learn test
            atol = 1e-1

        # Check accuracy consistency
        assert_allclose(lr_32.coef_, lr_32.coef_.astype(np.float32), atol=atol)

def test_warm_start_converge_LR():
    # Test to see that the logistic regression converges on warm start,
    # with multi_class='multinomial'. Non-regressive test for #10836

    rng = np.random.RandomState(0)
    X = np.concatenate((rng.randn(100, 2) + [1, 1], rng.randn(100, 2)))
    y = np.array([1] * 100 + [-1] * 100)
    lr_no_ws = LogisticRegression(
        multi_class="multinomial", solver="qning-ista", warm_start=False, random_state=0
    )
    lr_ws = LogisticRegression(
        multi_class="multinomial", solver="qning-ista", warm_start=True, random_state=0
    )

    lr_no_ws_loss = log_loss(y, lr_no_ws.fit(X, y).predict_proba(X))
    for i in range(10):
        lr_ws.fit(X, y)
    lr_ws_loss = log_loss(y, lr_ws.predict_proba(X))
    assert_allclose(lr_no_ws_loss, lr_ws_loss, rtol=1e-3)


def test_elastic_net_coeffs():
    # make sure elasticnet penalty gives different coefficients from l1 and l2
    # with saga solver (l1_ratio different from 0 or 1)
    X, y = make_classification(random_state=0)

    alpha = 2
    n_samples = 100
    lambda_1 = 1 / (n_samples * alpha)
    lambda_2 = 1 / (n_samples * alpha)
    coeffs = list()
    for penalty in ("l1", "l2"):
        if penalty in ["l1", "l2"]:
            lambda_2 = 0
        lr = LogisticRegression(
            penalty=penalty, lambda_1=lambda_1, solver="qning-ista", random_state=0, lambda_2=lambda_2
        )
        lr.fit(X, y)
        coeffs.append(lr.coef_)

    l1_coeffs, l2_coeffs = coeffs
    # make sure coeffs differ by at least .1
    assert not np.allclose(l2_coeffs, l1_coeffs, rtol=0, atol=0.1)

@pytest.mark.parametrize("C", np.logspace(-2, 2, 4))
def test_l1_versus_sgd(C):
    # Compare elasticnet penalty in LogisticRegression() and SGD(loss='log')
    n_samples = 500
    X, y = make_classification(
        n_samples=n_samples,
        n_classes=2,
        n_features=5,
        n_informative=5,
        n_redundant=0,
        n_repeated=0,
        random_state=1,
    )
    X = scale(X)

    sgd = SGDClassifier(
        penalty="l1",
        random_state=1,
        fit_intercept=False,
        tol=None,
        max_iter=2000,
        alpha=1 / (C * n_samples),
        loss="log_loss",
    )
    log = LogisticRegression(
        penalty="l1",
        random_state=1,
        fit_intercept=False,
        tol=1e-12,
        max_iter=6000,
        lambda_1=1 / (C * n_samples),
        solver="qning-ista",
    )

    sgd.fit(X, y)
    log.fit(X, y)
    assert_array_almost_equal(np.squeeze(sgd.coef_), np.transpose(log.coef_), decimal=1)

@pytest.mark.parametrize(
    "est",
    [
        LogisticRegression(random_state=0, max_iter=500),
    ],
    ids=lambda x: x.__class__.__name__,
)
@pytest.mark.parametrize("solver", solvers)
def test_logistic_regression_multi_class_auto(est, solver):
    # check multi_class='auto' => multi_class='ovr' iff binary y or liblinear

    def fit(X, y, **kw):
        return clone(est).set_params(**kw).fit(X, y)

    scaled_data = scale(iris.data)
    X = scaled_data[::10]
    X2 = scaled_data[1::10]
    y_multi = iris.target[::10].astype(np.int32)
    
    y_bin = y_multi == 0
    est_auto_bin = fit(X, y_bin, multi_class="auto", solver=solver)
    est_ovr_bin = fit(X, y_bin, multi_class="ovr", solver=solver)
    assert_allclose(est_auto_bin.coef_, est_ovr_bin.coef_)
    assert_allclose(est_auto_bin.predict_proba(X2), est_ovr_bin.predict_proba(X2))

    est_auto_multi = fit(X, y_multi, multi_class="auto", solver=solver)
    est_multi_multi = fit(X, y_multi, multi_class="multinomial", solver=solver)
    assert_allclose(est_auto_multi.coef_, est_multi_multi.coef_)
    assert_allclose(
        est_auto_multi.predict_proba(X2), est_multi_multi.predict_proba(X2)
    )

    # Make sure multi_class='ovr' and ='multinomial' should be the same in the binary case
    assert np.allclose(
        est_auto_bin.coef_,
        fit(X, y_bin, multi_class="multinomial", solver=solver).coef_,
    )

    # Make sure multi_class='ovr' is distinct from ='multinomial'
    assert not np.allclose(
        est_auto_multi.coef_,
        fit(X, y_multi, multi_class="ovr", solver=solver).coef_,
    )



@pytest.mark.parametrize("solver", solvers)
def test_penalty_none(solver):
    # - Make sure warning is raised if penalty='none' and lambda_1 is set to a
    #   non-default value.
    # - Make sure setting penalty='none' is equivalent to setting lambda_1=np.inf with
    #   l2 penalty.
    X, y = make_classification(n_samples=1000, random_state=0)

    msg = "Setting penalty='none' will ignore the lambda_1"
    lr = LogisticRegression(penalty="none", solver=solver, lambda_1=4)
    with pytest.warns(UserWarning, match=msg):
        lr.fit(X, y)

    lr_none = LogisticRegression(penalty="none", solver=solver, random_state=0, tol=10e-15)
    lr_l2_C_inf = LogisticRegression(
        penalty="l2", lambda_1=1/np.inf, solver=solver, random_state=0, tol=10e-15
    )
    pred_none = lr_none.fit(X, y).predict(X)
    pred_l2_C_inf = lr_l2_C_inf.fit(X, y).predict(X)

    assert_array_equal(pred_none, pred_l2_C_inf)



import torch

from cyanure_gpu.losses.loss import LinearLossVec

from cyanure_gpu.logger import setup_custom_logger

logger = setup_custom_logger("INFO")


class SquareLoss(LinearLossVec):

    def __init__(self, data: torch.Tensor, y: torch.Tensor, intercept: bool):
        super().__init__(data, y, intercept)
        self.id = "SQUARE"

    def pre_compute(self, input: torch.Tensor) -> float:

        tmp = self.pred_tensor(input, None)

        tmp = torch.sub(tmp, self.labels)

        return tmp

    def eval_tensor(self, input: torch.Tensor, matmul_result: torch.Tensor = None, precompute: torch.Tensor = None) -> float:
        if precompute is None:
            if matmul_result is not None:
                grad1 = matmul_result
            else:
                grad1 = self.pred_tensor(input, None)
            grad1 = torch.sub(grad1, self.labels)
        else:
            grad1 = precompute
        return 0.5*torch.linalg.norm(grad1)**2/grad1.size(dim=0)

    def eval(self, input: torch.Tensor, i: int) -> float:
        res = self.labels[i] - self.pred(i, input)
        return 0.5*res*res

    def print(self) -> None:
        logger.info("Square Loss is used")

    def fenchel(self, input: torch.Tensor) -> float:
        return 0.5*torch.linalg.norm(input)**2/input.size(dim=0)+torch.dot(input, self.labels)/input.size(dim=0)

    def scal_grad(self, input: torch.Tensor, i: int) -> float:
        return self.pred(i, input, None) - self.labels[i]

    def get_grad_aux(self, input: torch.Tensor, matmul_result: torch.Tensor = None,
                     precompute: torch.Tensor = None) -> torch.Tensor:

        if precompute is None:
            if matmul_result is not None:
                grad1 = matmul_result
            else:
                grad1 = self.pred_tensor(input, None)
            grad1 = torch.sub(grad1, self.labels)
        else:
            grad1 = precompute

        return grad1

    def lipschitz_constant(self) -> float:
        return 1.0

    def get_dual_constraints(self, grad1: torch.Tensor) -> torch.Tensor:
        if (self.intercept):
            grad1 = grad1 - torch.mean(grad1)

        return grad1
